"""Common front-end to all ingestion tools.
"""
__copyright__ = "Copyright (C) 2016,2018  Martin Blais"
__license__ = "GNU GPLv2"

from os import path
import logging
import os
import re
import runpy
import stat
import sys
import unittest
import warnings

from beancount.ingest import importer
from beancount.ingest import cache
from beancount.utils import test_utils
from beancount.parser import version
from beancount.ingest import identify
from beancount.ingest import extract
from beancount.ingest import file


DESCRIPTION = ("Identify, extract or file away data downloaded from "
               "financial institutions.")


def identify_main():
    """Main program for 'identify' function."""
    return trampoline_to_ingest(identify)


def extract_main():
    """Main program for 'extract' function."""
    return trampoline_to_ingest(extract)


def file_main():
    """Main program for 'file' function."""
    return trampoline_to_ingest(file)


def ingest(importers_list, detect_duplicates_func=None, hooks=None):
    """Driver function that calls all the ingestion tools.

    Put a call to this function at the end of your importer configuration to
    make your import script; this should be its main function, like this:

      from beancount.ingest.scripts_utils import ingest
      my_importers = [ ... ]
      ingest(my_importers)

    This more explicit way of invoking the ingestion is now the preferred way to
    invoke the various tools, and replaces calling the bean-identify,
    bean-extract, bean-file tools with a --config argument. When you call the
    import script itself (as as program) it will parse the arguments, expecting
    a subcommand ('identify', 'extract' or 'file') and corresponding
    subcommand-specific arguments.

    Here you can override some importer values, such as installing a custom
    duplicate finding hook, and eventually more. Note that this newer invocation
    method is optional and if it is not present, a call to ingest() is generated
    implicitly, and it functions as it used to. Future configurable
    customization of the ingestion process will be implemented by inserting new
    arguments to this function, this is the motivation behind doing this.

    Note that invocation by the three bean-* ingestion tools is still supported,
    and calling ingest() explicitly from your import configuration file will not
    break these tools either, if you invoke them on it; the values you provide
    to this function will be used by those tools.

    Args:
      importers_list: A list of importer instances. This is used as a
        chain-of-responsibility, called on each file.
      detect_duplicates_func: (DEPRECATED) An optional function which accepts a
        list of lists of imported entries and a list of entries already existing
        in the user's ledger. See function find_duplicate_entries(), which is
        the default implementation for this. Use 'filter_funcs' instead.
      hooks: An optional list of hook functions to apply to the list of extract
        (filename, entries) pairs, in order. This replaces
        'detect_duplicates_func'.
    """
    if detect_duplicates_func is not None:
        warnings.warn("Argument 'detect_duplicates_func' is deprecated.")
        # Fold it in hooks.
        if hooks is None:
            hooks = []
        hooks.insert(0, detect_duplicates_func)
        del detect_duplicates_func

    if ingest.args is not None:
        # The script has been called from one of the bean-* ingestion tools.
        # 'ingest.args' is only set when we're being invoked from one of the
        # bean-xxx tools (see below).

        # Mark this function as called, so that if it is called from an import
        # triggered by one of the ingestion tools, it won't be called again
        # afterwards.
        ingest.was_called = True

        # Use those args rather than to try to parse the command-line arguments
        # from a naked ingest() call as a script. {39c7af4f6af5}
        args, parser = ingest.args
    else:
        # The script is called directly. This is the main program of the import
        # script itself. This is the new invocation method.
        parser = version.ArgumentParser(description=DESCRIPTION)

        # Use required on subparsers.
        # FIXME: Remove this when we require version 3.7 or above.
        kwargs = {}
        if sys.version_info >= (3, 7):
            kwargs['required'] = True
        subparsers = parser.add_subparsers(dest='command', **kwargs)

        parser.add_argument('--downloads', '-d', metavar='DIR-OR-FILE',
                            action='append', default=[],
                            help='Filenames or directories to search for files to import')

        for cmdname, module in [('identify', identify),
                                ('extract', extract),
                                ('file', file)]:
            parser_cmd = subparsers.add_parser(cmdname, help=module.DESCRIPTION)
            parser_cmd.set_defaults(command=module.run)
            module.add_arguments(parser_cmd)

        args = parser.parse_args()

        if not args.downloads:
            args.downloads.append(os.getcwd())

        # Implement required ourselves.
        # FIXME: Remove this when we require version 3.7 or above.
        if not (sys.version_info >= (3, 7)):
            if not hasattr(args, 'command'):
                parser.error("Subcommand is required.")

    abs_downloads = list(map(path.abspath, args.downloads))
    args.command(args, parser, importers_list, abs_downloads, hooks=hooks)
    return 0


# A global sentinel to mark whether ingest() has been called at least once.
ingest.was_called = False

# A global value of program args for the ingest subcommand. If the command is
# being trampolined (that is, called by one of the bean-* ingestion tools), and
# the arguments from the bean-xxx tool have already been parsed, save them here
# for reuse by ingest() instead of attempting to convert the per-command
# arguments into generic ingest() arguments (which is impossible to do without
# parsing in the first place due to the support for --argument value).
ingest.args = None


def create_legacy_arguments_parser(description: str, run_func: callable):
    """Create an arguments parser for all the ingestion bean-tools.

    Args:
      description: The program description string.
      func: A callable function to run the particular command.
    Returns:
      An argparse.Namespace instance with the rest of arguments in 'rest'.
    """
    parser = version.ArgumentParser(description=description)

    parser.add_argument('config', action='store', metavar='CONFIG_FILENAME',
                        help=('Importer configuration file. '
                              'This is a Python file with a data structure that '
                              'is specific to your accounts'))

    parser.add_argument('downloads', nargs='+', metavar='DIR-OR-FILE',
                        default=[],
                        help='Filenames or directories to search for files to import')

    parser.set_defaults(command=run_func)

    return parser


def trampoline_to_ingest(module):
    """Parse arguments for bean tool, import config script and ingest.

    This function is called by the three bean-* tools to support the older
    import files, which only required a CONFIG object to be defined in them.

    Args:
      module: One of the identify, extract or file module objects.
    Returns:
      An execution return code.
    """
    # Disable debugging logging which is turned on by default in chardet.
    logging.getLogger('chardet.charsetprober').setLevel(logging.INFO)
    logging.getLogger('chardet.universaldetector').setLevel(logging.INFO)

    parser = create_legacy_arguments_parser(module.DESCRIPTION, module.run)
    module.add_arguments(parser)
    return run_import_script_and_ingest(parser)


def run_import_script_and_ingest(parser, argv=None, importers_attr_name='CONFIG'):
    """Run the import script and optionally call ingest().

    This path is only called when trampolined by one of the bean-* ingestion
    tools.

    Args:
      parser: The parser instance, used only to report errors.
      importers_attr_name: The name of the special attribute in the module which
        defines the importers list.
    Returns:
      An execution return code.
    """
    args = parser.parse_args(args=argv)

    # Check the existence of the config.
    if not path.exists(args.config) or path.isdir(args.config):
        parser.error("File does not exist: '{}'".format(args.config))

    # Check the existence of all specified files.
    for filename in args.downloads:
        if not path.exists(filename):
            parser.error("File does not exist: '{}'".format(filename))

    # Reset the state of ingest() being called (for unit tests, which use the
    # same runtime with run_with_args).
    ingest.was_called = False

    # Save the arguments parsed from the command-line as default for
    # {39c7af4f6af5}.
    ingest.args = args, parser

    # Evaluate the importer script/module.
    mod = runpy.run_path(args.config)

    # If the importer script has already called ingest() within itself, don't
    # call it again. We're done. This allows the use to insert an explicit call
    # to ingest() while still running the bean-* ingestion tools on the file.
    if ingest.was_called:
        return 0

    # ingest() hasn't been called by the script so we assume it isn't
    # present in it. So we now run the ingestion by ourselves here, without
    # specifying any of the newer optional arguments.
    importers_list = mod[importers_attr_name]
    return ingest(importers_list)


class _TestFileImporter(importer.ImporterProtocol):

    def __init__(self, name, account, regexp_mime, regexp_contents):
        self._name = name
        self.account = account
        self.regexp_mime = regexp_mime
        self.regexp_contents = regexp_contents

    def name(self):
        return self._name

    def identify(self, file):
        mimetype = file.convert(cache.mimetype)
        if re.match(self.regexp_mime, mimetype):
            return True
        if re.search(self.regexp_contents, file.contents()):
            return True
        return False

    def file_account(self, _):
        return self.account


IMPORT_FILE = """\
#!/usr/bin/env python3
from beancount.ingest import scripts_utils

CONFIG = [
    scripts_utils._TestFileImporter(
        'mybank-checking-ofx', 'Assets:Checking',
        'application/x-ofx', '<FID>3011'),
    scripts_utils._TestFileImporter(
        'mybank-credit-csv', 'Liabilities:CreditCard',
        'text/csv', '.*DATE,TRANSACTION ID,DESCRIPTION,QUANTITY,SYMBOL'),
]
"""

INGEST_MAIN = """\
scripts_utils.ingest(CONFIG)
"""

OFX_FILE = """\
OFXHEADER:100
DATA:OFXSGML
VERSION:102
SECURITY:NONE
ENCODING:USASCII
CHARSET:1252
COMPRESSION:NONE
OLDFILEUID:NONE
NEWFILEUID:NONE

<OFX><SIGNONMSGSRSV1><SONRS><STATUS><CODE>0<SEVERITY>INFO<MESSAGE>Login successful</STATUS><DTSERVER>20131217204544.559[-7:MST]<LANGUAGE>ENG<FI><ORG>OFCT<FID>3011</FI><ORIGIN.ID>FMPWeb<START.TIME>20131217204544</SONRS></SIGNONMSGSRSV1><CREDITCARDMSGSRSV1><CCSTMTTRNRS><TRNUID>0<STATUS><CODE>0<SEVERITY>INFO</STATUS><CCSTMTRS><CURDEF>USD<CCACCTFROM><ACCTID>092243467384967<DOWNLOAD.FLAG>false<DOWNLOAD.TYPE>downloadSince<AMEX.BASICACCT>090341355486768<DAYS.SINCE>true<AMEX.ROLE>B<AMEX.UNIVID>iHJPMCPVMUZESUMTMIASKPSHBZOJZQMZ</CCACCTFROM><BANKTRANLIST><DTSTART>20131213050000.000[-7:MST]<DTEND>20131217050000.000[-7:MST]<STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131206000000.000[-7:MST]<DTUSER>20131206000000.000[-7:MST]<TRNAMT>-75<FITID>132124581254980455<REFNUM>140941621247980353<NAME>Cvzndybfhlgsy Kbptkt010-743-2492<MEMO>87278814438304-062-9392</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131208000000.000[-7:MST]<DTUSER>20131207000000.000[-7:MST]<TRNAMT>-29.5<FITID>139251640671720832<REFNUM>411944529384600439<NAME>YJTEJSYC JXJ 38137 80223112202<MEMO>841814901332133213240</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131208000000.000[-7:MST]<DTUSER>20131208000000.000[-7:MST]<TRNAMT>-96.73<FITID>518223640481029842<REFNUM>349922421383839452<NAME>TEMSRB TQBHHWZO CZYKCGDX.LAR/CD<MEMO>BVR5D49Q7S3 IWOUXSFCCIZ</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131209000000.000[-7:MST]<DTUSER>20131208000000.000[-7:MST]<TRNAMT>-45.49<FITID>410313240598642566<REFNUM>201153532386740368<NAME>JWNNJ VPVHHV -  HWKZIGH QXWR   <MEMO>35905       UJGZDQD IUTFL</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131209000000.000[-7:MST]<DTUSER>20131208000000.000[-7:MST]<TRNAMT>-01.7<FITID>118954331459601590<REFNUM>112944250496740196<NAME>ZPIKRWGV EBQVUE 4521XJT AYDM   <MEMO>227092130   2924489277</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131209000000.000[-7:MST]<DTUSER>20131208000000.000[-7:MST]<TRNAMT>-30.9<FITID>118335238578609388<REFNUM>542324610398801568<NAME>SBMUZYXG XRB 98038 03324302420<MEMO>853700608200014392232</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131209000000.000[-7:MST]<DTUSER>20131206000000.000[-7:MST]<TRNAMT>-39.72<FITID>141044448255701269<REFNUM>230245232285603469<NAME>SITEH NIHOX HTYZBAWP392-139-734<MEMO>84165624260 423-151-8227</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131209000000.000[-7:MST]<DTUSER>20131208000000.000[-7:MST]<TRNAMT>-22.09<FITID>111921351569432591<REFNUM>101153620388713392<NAME>HGJQJEOB PCQ 08418 1KVO VVDJ   <MEMO>84408170244 1144012409</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131211000000.000[-7:MST]<DTUSER>20131210000000.000[-7:MST]<TRNAMT>-22.14<FITID>548935642111458816<REFNUM>328141439181292814<NAME>LILFN DVRIFI -  LJBPFDT HYSF   <MEMO>32793       IBTSNAF UDOMK</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131211000000.000[-7:MST]<DTUSER>20131210000000.000[-7:MST]<TRNAMT>-12.67<FITID>330054241010450007<REFNUM>342912468199362629<NAME>PEHKXNPZ PNW 91458 21015119128<MEMO>823891919293222482430</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131214000000.000[-7:MST]<DTUSER>20131212000000.000[-7:MST]<TRNAMT>-31<FITID>402252668937162476<REFNUM>448222302958184678<NAME>YHVWV NNPYW HRQZDUOJ201-251-533<MEMO>65093823538 334-432-6338</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131214000000.000[-7:MST]<DTUSER>20131213000000.000[-7:MST]<TRNAMT>-64.25<FITID>542124402167304547<REFNUM>222333308839462735<NAME>MRKW'G #814 HYLDQF OVNN PSCS   <MEMO>89318083925 WOQO'G</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131216000000.000[-7:MST]<DTUSER>20131215000000.000[-7:MST]<TRNAMT>-42.01<FITID>210943512087240724<REFNUM>501152389299358014<NAME>ICPBVFY #2321 929883SXU QRKN   <MEMO>12054513980 3062749128</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131216000000.000[-7:MST]<DTUSER>20131215000000.000[-7:MST]<TRNAMT>-5<FITID>310215309277250199<REFNUM>448945618978219897<NAME>LON BXBZA 409 QEPNUAMCO WDOD   <MEMO>31345520940 4169979285</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131216000000.000[-7:MST]<DTUSER>20131215000000.000[-7:MST]<TRNAMT>-0.69<FITID>430314310166238818<REFNUM>109924680968112897<NAME>CFXUF LGTXVL -  VXCJCNI EBUH   <MEMO>308496      NSIWFWL RHSFP</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131216000000.000[-7:MST]<DTUSER>20131214000000.000[-7:MST]<TRNAMT>-34.43<FITID>330252321978428019<REFNUM>511914698876210997<NAME>DGCV GXD MJKRYD     JLC IYLM   <MEMO>67935504527 134-064-6852</STMTTRN><STMTTRN><TRNTYPE>DEBIT<DTPOSTED>20131215000000.000[-7:MST]<DTUSER>20131214000000.000[-7:MST]<TRNAMT>-24<FITID>149941491182603315<REFNUM>429354298891572407<NAME>UPTMFSAD DSD 12354 14003323410<MEMO>963402919381906089541</STMTTRN></BANKTRANLIST><LEDGERBAL><BALAMT>-3609.07<DTASOF>20131217050000.000[-7:MST]</LEDGERBAL><CYCLECUT.INDICATOR>false<PURGE.INDICATOR>false<INTL.INDICATOR>false</CCSTMTRS></CCSTMTTRNRS></CREDITCARDMSGSRSV1></OFX>
"""

CSV_FILE = """\
DATE,TRANSACTION ID,DESCRIPTION,QUANTITY,SYMBOL,PRICE,COMMISSION,AMOUNT,NET CASH BALANCE,REG FEE,SHORT-TERM RDM FEE,FUND REDEMPTION FEE, DEFERRED SALES CHARGE
07/02/2013,10223506553,ORDINARY DIVIDEND (HDV),,HDV,,,31.04,31.04,,,,
07/02/2013,10224851005,MONEY MARKET PURCHASE,,,,,-31.04,0.00,,,,
07/02/2013,10224851017,MONEY MARKET PURCHASE (MMDA10),31.04,MMDA10,,,0.00,0.00,,,,
09/30/2013,10561187188,ORDINARY DIVIDEND (HDV),,HDV,,,31.19,31.19,,,,
09/30/2013,10563719172,MONEY MARKET PURCHASE,,,,,-31.19,0.00,,,,
09/30/2013,10563719198,MONEY MARKET PURCHASE (MMDA10),31.19,MMDA10,,,0.00,0.00,,,,
***END OF FILE***
"""

TXT_FILE = """\
Some random text file.
"""

class TestScriptsBase(test_utils.TestTempdirMixin, unittest.TestCase):

    FILES = {
        # Old style configuration which provides importers as a module attribute.
        'test.import': IMPORT_FILE,
        # New style configuration, which is just a script calling ingest() itself.
        'testimport.py': IMPORT_FILE + INGEST_MAIN,
        # Example input files.
        'Downloads/ofxdownload.ofx': OFX_FILE,
        'Downloads/Subdir/bank.csv': CSV_FILE,
        'Downloads/Subdir/readme.txt': TXT_FILE,
        }

    def setUp(self):
        super().setUp()
        for filename, contents in self.FILES.items():
            absname = path.join(self.tempdir, filename)
            os.makedirs(path.dirname(absname), exist_ok=True)
            with open(absname, 'w') as file:
                file.write(contents)
            if filename.endswith('.py') or filename.endswith('.sh'):
                os.chmod(absname, stat.S_IRUSR|stat.S_IXUSR)



"""Unit tests for conversion functions.
"""

__copyright__ = "Copyright (C) 2014-2017  Martin Blais"
__license__ = "GNU GPLv2"

import datetime
import unittest

from beancount.core.number import MISSING
from beancount.core.number import D
from beancount.core.amount import A
from beancount.core.data import create_simple_posting as P
from beancount.core.data import create_simple_posting_with_cost as PCost
from beancount.core.position import Cost
from beancount.core.position import Position
from beancount.core import convert
from beancount.core import inventory
from beancount.core import prices
from beancount.core import data
from beancount import loader


def build_price_map_util(date_currency_price_tuples):
    """Build a partial price-map just for testing.

    Args:
      date_currency_price_tuples: A list of (datetime.date, currency-string,
        price-Amount) tuples to fill in the database with.
    Returns:
      A price_map, as per build_price_map().
    """
    return prices.build_price_map(
        [data.Price(None, date, currency, price)
         for date, currency, price in date_currency_price_tuples])


class TestPositionConversions(unittest.TestCase):
    """Test conversions to units, cost, weight and market-value for Position objects."""

    def _pos(self, units, cost=None, price=None):
        # Note: 'price' is only used in Posting class which derives from this test.
        self.assertFalse(price)
        return Position(units, cost)

    #
    # Units
    #

    def test_units(self):
        units = A("100 HOOL")
        self.assertEqual(units, convert.get_units(
            self._pos(units,
                      Cost(D("514.00"), "USD", None, None))))
        self.assertEqual(units, convert.get_units(
            self._pos(units, None)))

    #
    # Cost
    #

    def test_cost__empty(self):
        self.assertEqual(A("100 HOOL"), convert.get_cost(
            self._pos(A("100 HOOL"), None)))

    def test_cost__not_empty(self):
        self.assertEqual(A("51400.00 USD"), convert.get_cost(
            self._pos(A("100 HOOL"),
                      Cost(D("514.00"), "USD", None, None))))

    def test_cost__missing(self):
        self.assertEqual(A("100 HOOL"), convert.get_cost(
            self._pos(A("100 HOOL"),
                      Cost(MISSING, "USD", None, None))))

    #
    # Weight
    #

    def test_weight__no_cost(self):
        self.assertEqual(A("100 HOOL"), convert.get_weight(
            self._pos(A("100 HOOL"), None)))

    def test_weight__with_cost(self):
        self.assertEqual(A("51400.00 USD"), convert.get_weight(
            self._pos(A("100 HOOL"),
                      Cost(D("514.00"), "USD", None, None))))

    def test_weight__with_cost_missing(self):
        self.assertEqual(A("100 HOOL"), convert.get_weight(
            self._pos(A("100 HOOL"),
                      Cost(MISSING, "USD", None, None))))

    def test_old_test(self):
        # Entry without cost, without price.
        posting = P(None, "Assets:Bank:Checking", "105.50", "USD")
        self.assertEqual(A("105.50 USD"),
                         convert.get_weight(posting))

        # Entry without cost, with price.
        posting = posting._replace(price=A("0.90 CAD"))
        self.assertEqual(A("94.95 CAD"),
                         convert.get_weight(posting))

        # Entry with cost, without price.
        posting = PCost(None, "Assets:Bank:Checking", "105.50", "USD", "0.80", "EUR")
        self.assertEqual(A("84.40 EUR"),
                         convert.get_weight(posting))

        # Entry with cost, and with price (the price should be ignored).
        posting = posting._replace(price=A("2.00 CAD"))
        self.assertEqual(A("84.40 EUR"),
                         convert.get_weight(posting))


    #
    # Value
    #

    PRICE_MAP_EMPTY = build_price_map_util([])

    PRICE_MAP_HIT = build_price_map_util([
        (datetime.date(2016, 2, 1), "HOOL", A("530.00 USD")),
        (datetime.date(2016, 2, 1), "USD", A("1.2 CAD")),
    ])


    def test_value__no_currency(self):
        pos = self._pos(A("100 HOOL"), None)
        self.assertEqual(A("100 HOOL"),
                         convert.get_value(pos, self.PRICE_MAP_EMPTY))
        self.assertEqual(A("100 HOOL"),
                         convert.get_value(pos, self.PRICE_MAP_HIT))

    def test_value__currency_from_cost(self):
        pos = self._pos(A("100 HOOL"), Cost(D("514.00"), "USD", None, None))
        self.assertEqual(A("53000.00 USD"),
                         convert.get_value(pos, self.PRICE_MAP_HIT))
        self.assertEqual(A("100 HOOL"),
                         convert.get_value(pos, self.PRICE_MAP_EMPTY))


    #
    # Conversion to another currency.
    #

    def test_convert_position__success(self):
        pos = self._pos(A("100 HOOL"), Cost(D("514.00"), "USD", None, None))
        self.assertEqual(A("53000.00 USD"),
                         convert.convert_position(pos, "USD", self.PRICE_MAP_HIT))

    def test_convert_position__miss_but_same_currency(self):
        pos = self._pos(A("100 HOOL"), Cost(D("514.00"), "USD", None, None))
        self.assertEqual(A("100 HOOL"),
                         convert.convert_position(pos, "USD", self.PRICE_MAP_EMPTY))

    def test_convert_position__miss_and_miss_rate_to_rate(self):
        pos = self._pos(A("100 HOOL"), Cost(D("514.00"), "USD", None, None))
        self.assertEqual(A("100 HOOL"),
                         convert.convert_position(pos, "JPY", self.PRICE_MAP_HIT))

    PRICE_MAP_RATEONLY = build_price_map_util([
        (datetime.date(2016, 2, 1), "USD", A("1.2 CAD")),
    ])

    def test_convert_position__miss_and_miss_value_rate(self):
        pos = self._pos(A("100 HOOL"), Cost(D("514.00"), "USD", None, None))
        self.assertEqual(A("100 HOOL"),
                         convert.convert_position(pos, "CAD", self.PRICE_MAP_RATEONLY))

    def test_convert_position__miss_and_miss_both(self):
        pos = self._pos(A("100 HOOL"), Cost(D("514.00"), "USD", None, None))
        self.assertEqual(A("100 HOOL"),
                         convert.convert_position(pos, "CAD", self.PRICE_MAP_EMPTY))

    def test_convert_position__miss_and_success_on_implieds(self):
        pos = self._pos(A("100 HOOL"), Cost(D("514.00"), "USD", None, None))
        self.assertEqual(A("63600.00 CAD"),
                         convert.convert_position(pos, "CAD", self.PRICE_MAP_HIT))

    #
    # Conversion of amounts to another currency.
    #

    def test_convert_amount__fail(self):
        amt = A("127.00 USD")
        self.assertEqual(amt,
                         convert.convert_amount(amt, "CAD", self.PRICE_MAP_EMPTY))

    def test_convert_amount__success(self):
        amt = A("127.00 USD")
        self.assertEqual(A("152.40 CAD"),
                         convert.convert_amount(amt, "CAD", self.PRICE_MAP_HIT))

    def test_convert_amount__noop(self):
        amt = A("127.00 USD")
        self.assertEqual(amt,
                         convert.convert_amount(amt, "USD", self.PRICE_MAP_EMPTY))
        self.assertEqual(amt,
                         convert.convert_amount(amt, "USD", self.PRICE_MAP_HIT))

    @loader.load_doc()
    def test_convert_amount_with_date(self, entries, _, __):
        """
        2013-01-01 price  USD  1.20 CAD
        2014-01-01 price  USD  1.25 CAD
        2015-01-01 price  USD  1.30 CAD
        """
        price_map = prices.build_price_map(entries)
        for date, exp_amount in [
                (None, A('130 CAD')),
                (datetime.date(2015, 1, 1), A('130 CAD')),
                (datetime.date(2014, 12, 31), A('125 CAD')),
                (datetime.date(2014, 1, 1), A('125 CAD')),
                (datetime.date(2013, 12, 31), A('120 CAD')),
                (datetime.date(2013, 1, 1), A('120 CAD')),
                (datetime.date(2012, 12, 31), A('100 USD')),
                ]:
            self.assertEqual(exp_amount,
                             convert.convert_amount(A('100 USD'), 'CAD', price_map, date))


class TestPostingConversions(TestPositionConversions):
    """Test conversions to units, cost, weight and market-value for Posting objects."""

    def _pos(self, units, cost=None, price=None):
        # Create a Posting instance instead of a Position.
        return data.Posting("Assets:AccountA", units, cost, price, None, None)

    def test_weight_with_cost_and_price(self):
        self.assertEqual(A("51400.00 USD"), convert.get_weight(
            self._pos(A("100 HOOL"),
                      Cost(D("514.00"), "USD", A("530.00 USD"), None))))

    def test_weight_with_only_price(self):
        self.assertEqual(A("53000.00 USD"), convert.get_weight(
            self._pos(A("100 HOOL"), None, A("530.00 USD"))))

    def test_value__currency_from_price(self):
        pos = self._pos(A("100 HOOL"), None, A("520.00 USD"))
        self.assertEqual(A("53000.00 USD"),
                         convert.get_value(pos, self.PRICE_MAP_HIT))
        self.assertEqual(A("100 HOOL"),
                         convert.get_value(pos, self.PRICE_MAP_EMPTY))

    def test_convert_position__currency_from_price(self):
        pos = self._pos(A("100 HOOL"), None, A("99999 USD"))
        self.assertEqual(A("63600.00 CAD"),
                         convert.convert_position(pos, "CAD", self.PRICE_MAP_HIT))


class TestMarketValue(unittest.TestCase):

    @loader.load_doc()
    def setUp(self, entries, _, __):
        """
        2013-06-01 price  USD  1.01 CAD
        2013-06-05 price  USD  1.05 CAD
        2013-06-06 price  USD  1.06 CAD
        2013-06-07 price  USD  1.07 CAD
        2013-06-10 price  USD  1.10 CAD

        2013-06-01 price  HOOL  101.00 USD
        2013-06-05 price  HOOL  105.00 USD
        2013-06-06 price  HOOL  106.00 USD
        2013-06-07 price  HOOL  107.00 USD
        2013-06-10 price  HOOL  110.00 USD

        2013-06-01 price  AAPL  91.00 USD
        2013-06-05 price  AAPL  95.00 USD
        2013-06-06 price  AAPL  96.00 USD
        2013-06-07 price  AAPL  97.00 USD
        2013-06-10 price  AAPL  90.00 USD
        """
        self.price_map = prices.build_price_map(entries)

    def test_no_change(self):
        balances = inventory.from_string('100 USD')
        market_value = balances.reduce(convert.get_value, self.price_map,
                                       datetime.date(2013, 6, 6))
        self.assertEqual(inventory.from_string('100 USD'), market_value)

    def test_other_currency(self):
        balances = inventory.from_string('100 CAD')
        market_value = balances.reduce(convert.get_value, self.price_map,
                                       datetime.date(2013, 6, 6))
        self.assertEqual(inventory.from_string('100 CAD'), market_value)

    def test_mixed_currencies(self):
        balances = inventory.from_string('100 USD, 90 CAD')
        market_value = balances.reduce(convert.get_value, self.price_map,
                                       datetime.date(2013, 6, 6))
        self.assertEqual(inventory.from_string('100 USD, 90 CAD'), market_value)

    def test_stock_single(self):
        balances = inventory.from_string('5 HOOL {0.01 USD}')
        market_value = balances.reduce(convert.get_value, self.price_map,
                                       datetime.date(2013, 6, 6))
        self.assertEqual(inventory.from_string('530 USD'), market_value)

    def test_stock_many_lots(self):
        balances = inventory.from_string('2 HOOL {0.01 USD}, 3 HOOL {0.02 USD}')
        market_value = balances.reduce(convert.get_value, self.price_map,
                                       datetime.date(2013, 6, 6))
        self.assertEqual(inventory.from_string('530 USD'), market_value)

    def test_stock_different_ones(self):
        balances = inventory.from_string('2 HOOL {0.01 USD}, 2 AAPL {0.02 USD}')
        market_value = balances.reduce(convert.get_value, self.price_map,
                                       datetime.date(2013, 6, 6))
        self.assertEqual(inventory.from_string('404 USD'), market_value)

    def test_stock_not_found(self):
        balances = inventory.from_string('2 MSFT {0.01 USD}')
        market_value = balances.reduce(convert.get_value, self.price_map,
                                       datetime.date(2013, 6, 6))
        self.assertEqual(inventory.from_string('2 MSFT'), market_value)


if __name__ == '__main__':
    unittest.main()



__copyright__ = "Copyright (C) 2015-2016  Martin Blais"
__license__ = "GNU GPLv2"

import datetime
import json
import textwrap
import unittest
from decimal import Decimal
from unittest import mock

from dateutil import tz
import requests

from beancount.prices.sources import yahoo
from beancount.core.number import D
from beancount.utils import date_utils



class MockResponse:
    """A mock requests.Models.Response object for testing."""

    def __init__(self, contents, status_code=requests.codes.ok):
        self.status_code = status_code
        self.contents = contents

    def json(self, **kwargs):
        return json.loads(self.contents, **kwargs)


class YahooFinancePriceFetcher(unittest.TestCase):

    def _test_get_latest_price(self):
        response = MockResponse(textwrap.dedent("""
            {"quoteResponse":
             {"error": null,
              "result": [{"esgPopulated": false,
                          "exchange": "TOR",
                          "exchangeDataDelayedBy": 15,
                          "exchangeTimezoneName": "America/Toronto",
                          "exchangeTimezoneShortName": "EDT",
                          "fullExchangeName": "Toronto",
                          "gmtOffSetMilliseconds": -14400000,
                          "language": "en-US",
                          "market": "ca_market",
                          "marketState": "CLOSED",
                          "quoteType": "ETF",
                          "regularMarketPrice": 29.99,
                          "regularMarketTime": 1522353589,
                          "sourceInterval": 15,
                          "symbol": "XSP.TO",
                          "tradeable": false}]}}
            """))
        with mock.patch('requests.get', return_value=response):
            srcprice = yahoo.Source().get_latest_price('XSP.TO')
        self.assertTrue(isinstance(srcprice.price, Decimal))
        self.assertEqual(D('29.99'), srcprice.price)
        timezone = datetime.timezone(datetime.timedelta(hours=-4), 'America/Toronto')
        self.assertEqual(datetime.datetime(2018, 3, 29, 15, 59, 49, tzinfo=timezone),
                         srcprice.time)
        self.assertEqual('CAD', srcprice.quote_currency)

    def test_get_latest_price(self):
        for tzname in "America/New_York", "Europe/Berlin", "Asia/Tokyo":
            with date_utils.intimezone(tzname):
                self._test_get_latest_price()

    def _test_get_historical_price(self):
        response = MockResponse(textwrap.dedent("""
            {"chart":
             {"error": null,
              "result": [{"indicators": {"adjclose": [{"adjclose": [29.236251831054688,
                                                                    29.16683006286621,
                                                                    29.196582794189453,
                                                                    29.226333618164062]}],
                                         "quote": [{"close": [29.479999542236328,
                                                              29.40999984741211,
                                                              29.440000534057617,
                                                              29.469999313354492],
                                                    "high": [29.510000228881836,
                                                             29.489999771118164,
                                                             29.469999313354492,
                                                             29.579999923706055],
                                                    "low": [29.34000015258789,
                                                            29.350000381469727,
                                                            29.399999618530273,
                                                            29.43000030517578],
                                                    "open": [29.360000610351562,
                                                             29.43000030517578,
                                                             29.43000030517578,
                                                             29.530000686645508],
                                                    "volume": [160800,
                                                               118700,
                                                               98500,
                                                               227800]}]},
                          "meta": {"chartPreviousClose": 29.25,
                                   "currency": "CAD",
                                   "currentTradingPeriod": {"post": {"end": 1522702800,
                                                                     "gmtoffset": -14400,
                                                                     "start": 1522699200,
                                                                     "timezone": "EDT"},
                                                            "pre": {"end": 1522675800,
                                                                    "gmtoffset": -14400,
                                                                    "start": 1522670400,
                                                                    "timezone": "EDT"},
                                                            "regular": {"end": 1522699200,
                                                                        "gmtoffset": -14400,
                                                                        "start": 1522675800,
                                                                        "timezone": "EDT"}},
                                   "dataGranularity": "1d",
                                   "exchangeName": "TOR",
                                   "exchangeTimezoneName": "America/Toronto",
                                   "firstTradeDate": 1018872000,
                                   "gmtoffset": -14400,
                                   "instrumentType": "ETF",
                                   "symbol": "XSP.TO",
                                   "timezone": "EDT",
                                   "validRanges": ["1d", "5d", "1mo", "3mo", "6mo", "1y",
                                                   "2y", "5y", "10y", "ytd", "max"]},
                          "timestamp": [1509111000,
                                        1509370200,
                                        1509456600,
                                        1509543000]}]}}"""))
        with mock.patch('requests.get', return_value=response):
            srcprice = yahoo.Source().get_historical_price(
                'XSP.TO', datetime.datetime(2017, 11, 1, 16, 0, 0, tzinfo=tz.tzutc()))
        self.assertTrue(isinstance(srcprice.price, Decimal))
        self.assertEqual(D('29.469999313354492'), srcprice.price)
        timezone = datetime.timezone(datetime.timedelta(hours=-4), 'America/Toronto')
        self.assertEqual(datetime.datetime(2017, 11, 1, 9, 30, tzinfo=timezone),
                         srcprice.time)
        self.assertEqual('CAD', srcprice.quote_currency)

    def test_get_historical_price(self):
        for tzname in "America/New_York", "Europe/Berlin", "Asia/Tokyo":
            with date_utils.intimezone(tzname):
                self._test_get_historical_price()

    def test_parse_response_error_status_code(self):
        response = MockResponse(
            '{"quoteResponse": {"error": "Not supported", "result": [{}]}}',
            status_code=400)
        with self.assertRaises(yahoo.YahooError):
            yahoo.parse_response(response)

    def test_parse_response_error_invalid_format(self):
        response = MockResponse(
            """{"quoteResponse": {"error": null, "result": [{}]},
             "chart": {"error": null, "result": [{}]}}""")
        with self.assertRaises(yahoo.YahooError):
            yahoo.parse_response(response)

    def test_parse_response_error_not_none(self):
        response = MockResponse(
            '{"quoteResponse": {"error": "Non-zero error", "result": [{}]}}')
        with self.assertRaises(yahoo.YahooError):
            yahoo.parse_response(response)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""A script to produce a document that describes assets and account information.

This script produces a document to be attached to a legal will, a document that
describes the full list of all assets and account names with institutions names
and addresses, to make the work of gathering and liquidating assets easy. It
uses various metadata fields to group Beancount accounts together and produce an
output that should be readable by a layperson.

A special field 'institution' is detected to identify root accounts that should
be included in the document rendered by this script.
"""
__copyright__ = "Copyright (C) 2015-2016  Martin Blais"
__license__ = "GNU GPLv2"

import collections
import logging
import io
import re
import sys

from beancount import loader
from beancount.core import realization
from beancount.core import getters
from beancount.core import account_types
from beancount.core import data
from beancount.core import account
from beancount.core import convert
from beancount.parser import options
from beancount.parser import version


def group_accounts_by_metadata(accounts_map, meta_name):
    """Group accounts by the value of a metadata field on its corresponding Open
    entry or in one of its parent accounts.

    Args:
      accounts_map: A mapping of account name to its Open entry.
      meta_name: A string, the name of a metadata key to extract to figure out the
        group name.
    Returns:
      A dict of group names (the values of the metadata field) to a list of account
      name strings, and a list of ignored accounts.
    """
    groups = collections.defaultdict(list)
    ignored_accounts = set()
    for account_ in accounts_map:
        # Find the group of this account; the group is defined as the first
        # parent account that has a particular metadata field. If an account is
        # not covered by a parent with the metadata, it defines its own group.
        for parent_account in account.parents(account_):
            open_entry = accounts_map.get(parent_account, None)
            if (open_entry and open_entry.meta and meta_name in open_entry.meta):
                group = open_entry.meta[meta_name]
                groups[group].append(account_)
                break
        else:
            ignored_accounts.add(account_)
    for group in groups.values():
        group.sort()
    return dict(groups), ignored_accounts


def find_institutions(entries, options_map):
    """Gather all the institutions and valid accounts from the list of entries.

    Args:
      entries: A list of entries.
      options_map: A dict of options, as per the parser.
    Returns:
      See group_accounts_by_metadata().
    """
    acc_types = options.get_account_types(options_map)

    # Filter out accounts that are closed or that are income accounts.
    open_close_map = getters.get_account_open_close(entries)
    accounts_map = {acc: open_entry
                    for acc, (open_entry, close_entry) in open_close_map.items()
                    if (account_types.is_balance_sheet_account(acc, acc_types) and
                        close_entry is None)}

    # Group the accounts using groups defined implicitly by metadata.
    return group_accounts_by_metadata(accounts_map, 'institution')


def get_first_meta(entries, field_name):
    """Get the first metadata value for a field in a list of entries.

    Args:
      entries: A list of directives, with metadata.
      field_name: A string, the key of a metadata field to fetch.
    Returns:
      The first seen value for field 'field_name'.
    """
    for entry in entries:
        if field_name in entry.meta:
            return entry.meta[field_name]


# Report data types.
Report = collections.namedtuple(
    'Report', 'title institutions')
InstitutionReport = collections.namedtuple(
    'Institution', 'name fields accounts')
AccountReport = collections.namedtuple(
    'Account', 'name open_date balance num_postings fields')


def create_report(entries, options_map):
    real_root = realization.realize(entries)

    # Find the institutions from the data.
    groups, ignored_accounts = find_institutions(entries, options_map)

    # List all the asset accounts which aren't included in the report.
    oc_map = getters.get_account_open_close(entries)
    open_map = {acc: open_entry for acc, (open_entry, _) in oc_map.items()}
    for acc in sorted(ignored_accounts):
        logging.info("Ignored account: %s", acc)

    # Gather missing fields and create a report object.
    institutions = []
    for name, accounts in sorted(groups.items()):
        # Get the institution fields, which is the union of the fields for all
        # the accounts with the institution fields.
        institution_accounts = [acc for acc in accounts
                                if 'institution' in open_map[acc].meta]

        institution_fields = {}
        for acc in institution_accounts:
            for key, value in open_map[acc].meta.items():
                institution_fields.setdefault(key, value)
        institution_fields.pop('filename', None)
        institution_fields.pop('lineno', None)

        # Create infos for each account in this institution.
        account_reports = []
        for acc in accounts:
            account_fields = {}
            for subacc in account.parents(acc):
                open_entry = open_map[subacc]
                if 'institution' in open_entry.meta:
                    break
                account_fields.update(open_entry.meta)
            account_fields.pop('filename', None)
            account_fields.pop('lineno', None)
            for field in institution_fields:
                account_fields.pop(field, None)

            real_node = realization.get(real_root, acc)
            account_reports.append(AccountReport(
                acc,
                open_entry.date,
                real_node.balance,
                sum(1 for posting in real_node.txn_postings
                    if isinstance(posting, data.TxnPosting)),
                account_fields))

        # Create the institution report.
        institution = InstitutionReport(name, institution_fields, account_reports)
        institutions.append(institution)

    return Report(options_map['title'], institutions)


XHTML_TEMPLATE_PRE = '''
<html>
  <head>
    <link href='https://fonts.googleapis.com/css?family=Open+Sans+Condensed:300'
          rel='stylesheet' type='text/css'>
    <link href='https://fonts.googleapis.com/css?family=Source+Sans+Pro'
          rel='stylesheet' type='text/css'>

    <style type="text/css">

body {
  font-family: 'Source Sans Pro', sans-serif;
}

h2 {
  margin-top: 2em;
}

table td {
  white-space: nowrap;
}

table td.name {
  width: 30%;
}
table td.open_date {
  width: 5%;
}

table.accounts thead td {
  font-weight: bold;
}

table.accounts td {
  background: #F0F0F0;
}

table.accounts {
  width: 100%;
  border-collapse: collapse;
}

table.accounts td {
  border: thin solid black;
  padding: 0.1em 0.3em;
}

td.balance {
  text-align: right;
}

table.fields {
  margin: 1em;
}

    </style>
  </head>
  <body>
'''
XHTML_TEMPLATE_POST = '''
  </body>
</html>
'''


def format_xhtml_report(report, options_map):
    oss = io.StringIO()
    oss.write(XHTML_TEMPLATE_PRE)

    oss.write('''
      <div class="report">
        <h1 class="title">{r.title}</h1>
    '''.format(r=report))

    for inst in report.institutions:
        oss.write('''
          <div class="institution">
            <h2>{i.name}</h2>
            {fields}
        '''.format(i=inst, fields=format_xhtml_table(sorted(inst.fields.items()))))

        # Compute the set of fields to render.
        unique_fields = {key
                         for acc in inst.accounts
                         for key in acc.fields}
        fieldnames = ['name', 'open_date'] + sorted(unique_fields) + ['balance']
        if inst.accounts:
            oss.write('''
              <table class="accounts">
              <thead>
              <tr>
            ''')
            for fieldname in fieldnames:
                oss.write('<th>{}</th>'.format(fieldname.strip('_').capitalize()))
            oss.write('</tr></thead>\n')

            for acc in inst.accounts:
                # Skip accounts without postings. This is the case, for
                # instance, of parent accounts only present in order to group
                # sub-accounts.
                if acc.num_postings == 0:
                    continue

                fields = acc.fields.copy()
                fields['name'] = acc.name
                fields['open_date'] = acc.open_date
                dcontext = options_map['dcontext']
                fields['balance'] = acc.balance.reduce(convert.get_cost).to_string(
                    dcontext.build(), False)
                oss.write('<tr>\n')
                for field in fieldnames:
                    oss.write('<td class="{}">{}</td>\n'.format(field,
                                                                fields.get(field, '')))
                oss.write('</tr>\n')
            oss.write('</table>\n')

        oss.write('</div>\n')
    oss.write('</div>\n')

    oss.write(XHTML_TEMPLATE_POST)
    return oss.getvalue()


def format_xhtml_table(items, klass='fields'):
    """Render a mapping of values to an HTML table.

    Args:
      items: A dict of key/value pairs.
    Returns:
      A string of rendered HTML.
    """
    oss = io.StringIO()
    oss.write('<table class="{}">\n'.format(klass))
    for key, value in items:
        if re.match('[a-z]+://', value):
            value = '<a href="{value}">{value}</a>'.format(value=value)
        oss.write('<tr><td>{}:</td><td>{}</td></tr>'.format(key.capitalize(), value))
    oss.write('</table>\n')
    return oss.getvalue()


def main():
    logging.basicConfig(level=logging.INFO, format='%(levelname)-8s: %(message)s')
    parser = version.ArgumentParser(description=__doc__.strip())
    parser.add_argument('filename', help='Beancount input filename')
    args = parser.parse_args()

    entries, _, options_map = loader.load_file(args.filename,
                                               log_errors=logging.error)

    report = create_report(entries, options_map)

    text = format_xhtml_report(report, options_map)
    sys.stdout.write(text)


if __name__ == '__main__':
    main()



__copyright__ = "Copyright (C) 2018  Martin Blais"
__license__ = "GNU GPLv2"

import unittest

from beancount.utils import import_utils
from beancount.utils import defdict


class TestImportSymbol(unittest.TestCase):

    def test_import_symbol(self):
        with self.assertRaises(ImportError):
            import_utils.import_symbol('beancount.nothing.data')
        with self.assertRaises((AttributeError, ModuleNotFoundError)):
            import_utils.import_symbol('beancount.utils.i_dont_exist')
        cls = import_utils.import_symbol('beancount.utils.defdict.DefaultDictWithKey')
        self.assertIs(defdict.DefaultDictWithKey, cls)


if __name__ == '__main__':
    unittest.main()



"""Beancount syntax parser.

IMPORTANT: The parser (and its grammar builder) produces "incomplete"
Transaction objects. This means that some of the data can be found missing from
the output of the parser and some of the data types vary slightly. Missing
components are replaced not by None, but by a special constant 'NA' which helps
diagnose problems if a user inadvertently attempts to work on an incomplete
posting instead of a complete one. Those incomplete entries are then run through
the "booking" routines which do two things simultaneously:

1. They find matching lots for reducing inventory positions, and
2. They interpolate missing numbers.

In doing so they normalize the entries to "complete" entries by converting a
position/lot's "cost" attribute from a CostSpec to a Cost. A Cost is similar to
an Amount in that it shares "number" and "currency" attributes, but also has a
label and a lot date. A CostSpec is similar to a Cost, but has all optional
data; it consists in a specification for matching against a particular inventory
lot.

Other parts of a posting may also be missing, not just parts of the cost.
Leaving out some parts of the input is used to invoke interpolation, to tell
Beancount to automatically compute the missing numbers (if possible).

Missing components will be set to the special value
"beancount.core.number.MISSING" until inventory booking and number interpolation
has been completed. The "MISSING" value should never appear in completed, loaded
transaction postings.

For instance, all of the units may be missing:

  INPUT: Assets:Account
  posting.units = MISSING

Or just the number of the units:

  INPUT: Assets:Account                    USD
  posting.units = Amount(MISSING, "USD")

You must always specify the currency.

If a price annotation is simply absent, it appears as None:

  INPUT: Assets:Account                 2 MXN
  posting.price = None

However, you may indicate that there is a price but have Beancount compute it
automatically:

  INPUT: Assets:Account                 2 MXN @
  posting.price = Amount(MISSING, MISSING)

Indicating the conversion currency is also possible (and recommended):

  INPUT: Assets:Account                 2 MXN @ USD
  posting.price = Amount(MISSING, "USD")

If a cost specification is provided, a "cost" attribute it set but it does not
refer to a Cost instance (as in complete entries) but rather to a CostSpec
instance. Some of the fields of a CostSpec may be MISSING if they were not
specified in the input. For example:

  INPUT: Assets:Account  1 HOOL {100 # 5 USD}
  posting.cost = CostSpec(Decimal("100"), Decimal("5"), "USD", None, None, False))

Note how we never consider the label of date override to be MISSING; this is
because those inputs are optional: A missing label is simply left unset in the
computed Cost, and a missing date override uses the date of the transaction
that contains the posting.

You can indicate that there is a total number to be filled in like this:

  INPUT: Assets:Account  1 HOOL {100 # USD}
  posting.cost = CostSpec(Decimal("100"), MISSING, "USD", None, None, False))

This is in contrast to the total value simple not being used:

  INPUT: Assets:Account  1 HOOL {100 USD}
  posting.cost = CostSpec(Decimal("100"), None, "USD", None, None, False))

Both per-unit and total numbers may be omitted as well, in which case, only the
number-per-unit portion of the CostSpec will appear as MISSING:

  INPUT: Assets:Account  1 HOOL {USD}
  posting.cost = CostSpec(MISSING, None, "USD", None, None, False))

And furthermore, all the cost basis may be missing:

  INPUT: Assets:Account  1 HOOL {}
  posting.cost = CostSpec(MISSING, None, MISSING, None, None, False))

If you ask for the lots to be merged, you get this:

  INPUT: Assets:Account  1 HOOL {*}
  posting.cost = CostSpec(MISSING, None, MISSING, None, None, True))

The numbers have to be computed by Beancount, so we output this with MISSING
values.

Of course, you can provide only the non-basis information, like just the date
or label:

  INPUT: Assets:Account  1 HOOL {2015-09-21}
  posting.cost = CostSpec(MISSING, None, MISSING, date(2015, 9, 21), None, True)

See the test beancount.parser.grammar_test.TestIncompleteInputs for examples and
corresponding expected values.
"""
__copyright__ = "Copyright (C) 2013-2016  Martin Blais"
__license__ = "GNU GPLv2"

import functools
import inspect
import textwrap
import io
import sys

from beancount.parser import _parser
from beancount.parser import grammar
from beancount.parser import printer
from beancount.parser import hashsrc
from beancount.core import data
from beancount.core.number import MISSING

from beancount.parser.grammar import ParserError
from beancount.parser.grammar import ParserSyntaxError
from beancount.parser.grammar import DeprecatedError


# disable pyflakes
# pylint: disable=pointless-statement
ParserError, ParserSyntaxError, DeprecatedError


# When importing the module, always check that the compiled source matched the
# installed source.
hashsrc.check_parser_source_files(_parser)


def is_posting_incomplete(posting):
    """Detect the presence of any elided amounts in a Posting.

    If any of the possible amounts are missing, this returns True.

    Args:
      entries: A directive.
    Returns:
      A boolean, true if there are some missing portions of any postings found.
    """
    units = posting.units
    if (units is MISSING or
        units.number is MISSING or
        units.currency is MISSING):
        return True
    price = posting.price
    if (price is MISSING or
        price is not None and (price.number is MISSING or
                               price.currency is MISSING)):
        return True
    cost = posting.cost
    if cost is not None and (cost.number_per is MISSING or
                             cost.number_total is MISSING or
                             cost.currency is MISSING):
        return True
    return False


def is_entry_incomplete(entry):
    """Detect the presence of elided amounts in Transactions.

    Args:
      entries: A directive.
    Returns:
      A boolean, true if there are some missing portions of any postings found.
    """
    if isinstance(entry, data.Transaction):
        if any(is_posting_incomplete(posting) for posting in entry.postings):
            return True
    return False


def parse_file(file, report_filename=None, report_firstline=0, **kw):
    """Parse a beancount input file and return Ledger with the list of
    transactions and tree of accounts.

    Args:
      file: file object or path to the file to be parsed.
      kw: a dict of keywords to be applied to the C parser.
    Returns:
      A tuple of (
        list of entries parsed in the file,
        list of errors that were encountered during parsing, and
        a dict of the option values that were parsed from the file.)
    """
    if file == '-':
        file = sys.stdin.buffer
    # It would be more appropriate here to check for io.RawIOBase but
    # that does not work for io.BytesIO despite it implementing the
    # readinto() method.
    elif not isinstance(file, io.IOBase):
        file = open(file, 'rb')
    builder = grammar.Builder()
    parser = _parser.Parser(builder)
    parser.parse(file, filename=report_filename, lineno=report_firstline, **kw)
    return builder.finalize()


def parse_string(string, report_filename=None, **kw):
    """Parse a beancount input file and return Ledger with the list of
    transactions and tree of accounts.

    Args:
      string: A string, the contents to be parsed instead of a file's.
      report_filename: A string, the source filename from which this string
        has been extracted, if any. This is stored in the metadata of the
        parsed entries.
      **kw: See parse.c. This function parses out 'dedent' which removes
        whitespace from the front of the text (default is False).
    Return:
      Same as the output of parse_file().
    """
    if kw.pop('dedent', None):
        string = textwrap.dedent(string)
    if isinstance(string, str):
        string = string.encode('utf8')
    if report_filename is None:
        report_filename = '<string>'
    file = io.BytesIO(string)
    return parse_file(file, report_filename=report_filename, **kw)


def parse_doc(expect_errors=False, allow_incomplete=False):
    """Factory of decorators that parse the function's docstring as an argument.

    Note that the decorators thus generated only run the parser on the tests,
    not the loader, so is no validation, balance checks, nor plugins applied to
    the parsed text.

    Args:
      expect_errors: A boolean or None, with the following semantics,
        True: Expect errors and fail if there are none.
        False: Expect no errors and fail if there are some.
        None: Do nothing, no check.
      allow_incomplete: A boolean, if true, allow incomplete input. Otherwise
        barf if the input would require interpolation. The default value is set
        not to allow it because we want to minimize the features tests depend on.
    Returns:
      A decorator for test functions.
    """
    def decorator(fun):
        """A decorator that parses the function's docstring as an argument.

        Args:
          fun: the function object to be decorated.
        Returns:
          A decorated test function.
        """
        filename = inspect.getfile(fun)
        lines, lineno = inspect.getsourcelines(fun)

        # decorator line + function definition line (I realize this is largely
        # imperfect, but it's only for reporting in our tests) - empty first line
        # stripped away.
        lineno += 1

        @functools.wraps(fun)
        def wrapper(self):
            assert fun.__doc__ is not None, (
                "You need to insert a docstring on {}".format(fun.__name__))
            entries, errors, options_map = parse_string(fun.__doc__,
                                                        report_filename=filename,
                                                        report_firstline=lineno,
                                                        dedent=True)

            if not allow_incomplete and any(is_entry_incomplete(entry)
                                            for entry in entries):
                self.fail("parse_doc() may not use interpolation.")

            if expect_errors is not None:
                if expect_errors is False and errors:
                    oss = io.StringIO()
                    printer.print_errors(errors, file=oss)
                    self.fail("Unexpected errors found:\n{}".format(oss.getvalue()))
                elif expect_errors is True and not errors:
                    self.fail("Expected errors, none found:")

            return fun(self, entries, errors, options_map)

        wrapper.__input__ = wrapper.__doc__
        wrapper.__doc__ = None
        return wrapper

    return decorator


def parse_many(string, level=0):
    """Parse a string with a snippet of Beancount input and replace vars from caller.

    Args:
      string: A string with some Beancount input.
      level: The number of extra stacks to ignore.
    Returns:
      A list of entries.
    Raises:
      AssertionError: If there are any errors.
    """
    # Get the locals in the stack for the callers and produce the final text.
    frame = inspect.stack()[level+1]
    varkwds = frame[0].f_locals
    input_string = textwrap.dedent(string.format(**varkwds))

    # Parse entries and check there are no errors.
    entries, errors, __ = parse_string(input_string)
    assert not errors

    return entries


def parse_one(string):
    """Parse a string with single Beancount directive and replace vars from caller.

    Args:
      string: A string with some Beancount input.
      level: The number of extra stacks to ignore.
    Returns:
      A list of entries.
    Raises:
      AssertionError: If there are any errors.
    """
    entries = parse_many(string, level=1)
    assert len(entries) == 1
    return entries[0]



__copyright__ = "Copyright (C) 2014-2017  Martin Blais"
__license__ = "GNU GPLv2"

import unittest

from beancount import loader
from beancount.parser import cmptest


class TestExampleExcludeTag(cmptest.TestCase):

    @loader.load_doc()
    def test_exclude_tag(self, entries, errors, __):
        """
            plugin "beancount.plugins.exclude_tag"

            2011-01-01 open Expenses:Restaurant
            2011-01-01 open Assets:Cash

            2011-05-17 * "This transaction should be included"
              Expenses:Restaurant   50.02 USD
              Assets:Cash

            2011-05-17 * "This transaction should be excluded" #virtual
              Expenses:Restaurant   11.11 USD
              Assets:Cash
        """
        self.assertFalse(errors)
        self.assertEqualEntries("""

            2011-01-01 open Expenses:Restaurant
            2011-01-01 open Assets:Cash

            2011-05-17 * "This transaction should be included"
              Expenses:Restaurant   50.02 USD
              Assets:Cash          -50.02 USD

        """, entries)


if __name__ == '__main__':
    unittest.main()



__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import os
import re
import textwrap
import tempfile
from os import path
import unittest

from beancount.parser import cmptest
from beancount.utils import test_utils
from beancount.scripts import doctor
from beancount.scripts import directories_test


class TestScriptDoctor(test_utils.TestCase):

    @test_utils.docfile
    def test_dump_lexer(self, filename):
        """
        2013-01-01 open Expenses:Restaurant
        2013-01-01 open Assets:Cash

        2014-03-02 * "Something"
          Expenses:Restaurant   50.02 USD
          Assets:Cash
        """
        with test_utils.capture() as stdout:
            test_utils.run_with_args(doctor.main, ['dump-lexer', filename])

        expected_output = """
            EOL               2 '\\n'
            DATE              2 '2013-01-01'
            OPEN              2 'open'
            ACCOUNT           2 'Expenses:Restaurant'
            EOL               3 '\\n'
            DATE              3 '2013-01-01'
            OPEN              3 'open'
            ACCOUNT           3 'Assets:Cash'
            EOL               4 '\\n'
            EOL               5 '\\n'
            DATE              5 '2014-03-02'
            ASTERISK          5 '*'
            STRING            5 '"'
            EOL               6 '\\n'
            INDENT            6 '  '
            ACCOUNT           6 'Expenses:Restaurant'
            NUMBER            6 '50.02'
            CURRENCY          6 'USD'
            EOL               7 '\\n'
            INDENT            7 '  '
            ACCOUNT           7 'Assets:Cash'
            EOL               8 '\\n'
            EOL               8 '\\x00'
        """
        self.assertLines(expected_output, stdout.getvalue())

    # pylint: disable=empty-docstring
    @test_utils.docfile
    def test_dump_lexer_empty(self, filename):
        ""
        with test_utils.capture():
            test_utils.run_with_args(doctor.main, ['dump-lexer', filename])

    @test_utils.docfile
    def test_dump_roundtrip(self, filename):
        """
        2013-01-01 open Expenses:Restaurant
        2013-01-01 open Assets:Cash

        2014-03-02 * "Something"
          Expenses:Restaurant   50.02 USD
          Assets:Cash
        """
        with test_utils.capture('stdout', 'stderr'):
            test_utils.run_with_args(doctor.main, ['roundtrip', filename])

    def test_list_options(self):
        with test_utils.capture():
            test_utils.run_with_args(doctor.main, ['list_options'])
            test_utils.run_with_args(doctor.main, ['list-options'])

    def test_deps(self):
        with test_utils.capture():
            test_utils.run_with_args(doctor.main, ['deps'])
            test_utils.run_with_args(doctor.main, ['checkdeps'])


class TestScriptCheckDirectories(directories_test.TestScriptCheckDirectories):

    @test_utils.docfile
    def test_invocation(self, filename):
        """
            2013-01-01 open Expenses:Restaurant
            2013-01-01 open Expenses:Movie
            2013-01-01 open Expenses:Alcohol
            2013-01-01 open Assets:Cash

            2014-03-02 * "Something"
              Expenses:Restaurant   50.02 USD
              Expenses:Alcohol      10.30 USD
              Expenses:Movie        25.00 USD
              Assets:Cash
        """
        for directory in self.TEST_DIRECTORIES:
            os.makedirs(path.join(self.tmpdir, directory))

        with test_utils.capture() as stdout:
            test_utils.run_with_args(doctor.main, ['directories', filename, self.tmpdir])
        self.assertEqual(2, len(stdout.getvalue().splitlines()))
        matches = set(match.group(1) for match in re.finditer("'(.*?)'", stdout.getvalue()))
        clean_matches = set(match[len(self.tmpdir)+1:]
                            if match.startswith(self.tmpdir)
                            else match
                            for match in matches)
        self.assertEqual({'Expenses/Restaurant/Sub',
                          'Expenses:Restaurant:Sub',
                          'Assets:Extra',
                          'Assets/Extra'}, clean_matches)


class TestScriptMissingOpen(cmptest.TestCase):

    @test_utils.docfile
    def test_missing_open(self, filename):
        """
            2013-01-01 open Expenses:Movie
            2013-01-01 open Assets:Cash

            2014-03-03 * "Something"
              Expenses:Restaurant   50.02 USD
              Expenses:Movie        25.00 USD
              Assets:Cash

            2014-04-04 * "Something"
              Expenses:Alcohol      10.30 USD
              Expenses:Movie        25.00 USD
              Assets:Cash
        """
        with test_utils.capture() as stdout:
            test_utils.run_with_args(doctor.main, ['missing-open', filename])

        self.assertEqualEntries("""

            2014-03-03 open Expenses:Restaurant
            2014-04-04 open Expenses:Alcohol

        """, stdout.getvalue())


class TestScriptDisplayContext(cmptest.TestCase):

    @test_utils.docfile
    def test_display_context(self, filename):
        """
            2013-01-01 open Expenses:Movie
            2013-01-01 open Assets:Cash

            2014-03-03 * "Something"
              Expenses:Restaurant   50.02 USD
              Expenses:Movie        25.00 USD
              Assets:Cash

            2014-04-04 * "Something"
              Expenses:Alcohol      10.30 USD
              Expenses:Movie        25.00 USD
              Assets:Cash
        """
        with test_utils.capture() as stdout:
            test_utils.run_with_args(doctor.main, ['display-context', filename])
        # Note: This probably deserves a little more love.
        self.assertTrue(stdout.getvalue())


class TestScriptContextualCommands(cmptest.TestCase):

    @test_utils.docfile
    def test_context(self, filename):
        """
            2013-01-01 open Expenses:Movie
            2013-01-01 open Assets:Cash

            2014-03-03 * "Something"
              Expenses:Restaurant   50.02 USD
              Expenses:Movie        25.00 USD
              Assets:Cash

            2014-04-04 * "Something"
              Expenses:Alcohol      10.30 USD
              Expenses:Movie        25.00 USD
              Assets:Cash
        """
        with test_utils.capture() as stdout:
            test_utils.run_with_args(doctor.main, ['context', filename, '6'])
        self.assertRegex(stdout.getvalue(), 'Location:')
        self.assertRegex(stdout.getvalue(), '50.02')

    @test_utils.docfile
    def test_context_multiple_files(self, filename):
        """
            2013-01-01 open Expenses:Movie
            2013-01-01 open Assets:Cash

            2014-03-03 * "Something"
              Expenses:Restaurant   50.02 USD
              Expenses:Movie        25.00 USD
              Assets:Cash

            2014-04-04 * "Something"
              Expenses:Alcohol      10.30 USD
              Expenses:Movie        25.00 USD
              Assets:Cash
        """

        with tempfile.NamedTemporaryFile('w') as topfile:
            topfile.write(textwrap.dedent("""
                include "{}"
            """.format(filename)))
            topfile.flush()
            with test_utils.capture() as stdout:
                test_utils.run_with_args(doctor.main, ['context', topfile.name,
                                                       '{}:6'.format(filename)])
            self.assertRegex(stdout.getvalue(), 'Location:')
            self.assertRegex(stdout.getvalue(), '50.02')

    @test_utils.docfile
    def test_linked(self, filename):
        """
            2013-01-01 open Expenses:Movie
            2013-01-01 open Assets:Cash

            2014-03-03 * "Apples" ^abc
              Expenses:Restaurant   50.02 USD
              Expenses:Movie        25.00 USD
              Assets:Cash

            2014-04-04 * "Something"
              Expenses:Alcohol      10.30 USD
              Expenses:Movie        25.00 USD
              Assets:Cash

            2014-05-05 * "Oranges" ^abc
              Expenses:Alcohol      10.30 USD
              Expenses:Movie        25.00 USD
              Assets:Cash
        """
        with test_utils.capture() as stdout:
            test_utils.run_with_args(doctor.main, ['linked', filename, '6'])
        self.assertRegex(stdout.getvalue(), 'Apples')
        self.assertRegex(stdout.getvalue(), 'Oranges')
        self.assertEqual(2, len(list(re.finditer(r'/(tmp|var/folders)/.*:\d+:',
                                                 stdout.getvalue()))))


if __name__ == '__main__':
    unittest.main()



"""This plugin validates that there are no duplicate transactions.
"""
__copyright__ = "Copyright (C) 2014, 2016-2017  Martin Blais"
__license__ = "GNU GPLv2"

from beancount.core import compare

__plugins__ = ('validate_no_duplicates',)


def validate_no_duplicates(entries, unused_options_map):
    """Check that the entries are unique, by computing hashes.

    Args:
      entries: A list of directives.
      unused_options_map: An options map.
    Returns:
      A list of new errors, if any were found.
    """
    unused_hashes, errors = compare.hash_entries(entries, exclude_meta=True)
    return entries, errors



__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import unittest

from beancount.reports import html_formatter
from beancount.core import data
from beancount.core import inventory
from beancount.core import display_context


class TestHTMLFormatter(unittest.TestCase):

    def test_functions(self):
        # Just a smoke test.
        formatter = html_formatter.HTMLFormatter(display_context.DEFAULT_DISPLAY_CONTEXT)
        formatter.render_account('Assets:US:Bank:Checking')
        formatter.render_inventory(inventory.from_string('10 CAD, 2 HOOL {500 USD}'))
        formatter.render_context('2b4722c3f89f43841cacf16325c2')
        formatter.render_link('fc6189c48a53')
        formatter.render_doc('/path/to/my/document.pdf')
        formatter.render_event_type('location')
        formatter.render_commodity(('HOOL', 'USD'))
        formatter.render_source(data.new_metadata('/path/to/my/input.beancount', 17))

    def test_render_inventory(self):
        formatter = html_formatter.HTMLFormatter(display_context.DEFAULT_DISPLAY_CONTEXT)
        balance = inventory.Inventory()
        self.assertEqual('', formatter.render_inventory(balance))

        balance = inventory.Inventory.from_string('111 USD, 222 CAD, 3 HOOL {400 USD}')
        html_balance = formatter.render_inventory(balance)
        self.assertRegex(html_balance, r'\b111\b')
        self.assertRegex(html_balance, r'\bUSD\b')
        self.assertRegex(html_balance, r'\b222\b')
        self.assertRegex(html_balance, r'\bCAD\b')
        self.assertRegex(html_balance, r'\b3\b')
        self.assertRegex(html_balance, r'\bHOOL\b')
        self.assertRegex(html_balance, r'\b400\b')


if __name__ == '__main__':
    unittest.main()



"""Beancount syntax lexer.
"""
__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import collections
import datetime
import io
import re
from decimal import Decimal

from beancount.core.data import new_metadata
from beancount.core import account
from beancount.parser import _parser


LexerError = collections.namedtuple('LexerError', 'source message entry')


class LexBuilder:
    """A builder used only for building lexer objects.

    Attributes:
      long_string_maxlines_default: Number of lines for a string to trigger a
          warning. This is meant to help users detecting dangling quotes in
          their source.
    """
    # pylint: disable=invalid-name

    def __init__(self):
        # A mapping of all the accounts created.
        self.accounts = {}

        # A regexp for valid account names.
        self.account_regexp = re.compile(account.ACCOUNT_RE)

        # A regexp for valid numbers.
        self.number_regexp = re.compile(r'(\d+|\d{1,3}(,\d{3})+)(\.\d+)?$')

        # Errors that occurred during lexing and parsing.
        self.errors = []

        # Default number of lines in string literals.
        self.long_string_maxlines_default = 64

    def get_invalid_account(self):
        """Return the name of an invalid account placeholder.

        When an account name is not deemed a valid one, replace it by
        this account name. This can be overridden by the parser to
        take into account the options.

        Returns:
          A string, the name of the root/type for invalid account names.
        """
        return 'Equity:InvalidAccountName'

    # Note: We could simplify the code by removing this if we could find a good
    # way to have the lexer communicate the error contents to the parser.
    def build_lexer_error(self, filename, lineno, message, exc_type=None): # {0e31aeca3363}
        """Build a lexer error and appends it to the list of pending errors.

        Args:
          message: The message of the error.
          exc_type: An exception type, if an exception occurred.
        """
        if not isinstance(message, str):
            message = str(message)
        if exc_type is not None:
            message = '{}: {}'.format(exc_type.__name__, message)
        self.errors.append(
            LexerError(new_metadata(filename, lineno), message, None))

    def DATE(self, year, month, day):
        """Process a DATE token.

        Args:
          year: integer year.
          month: integer month.
          day: integer day
        Returns:
          A new datetime object.
        """
        return datetime.date(year, month, day)

    def ACCOUNT(self, account_name):
        """Process an ACCOUNT token.

        This function attempts to reuse an existing account if one exists,
        otherwise creates one on-demand.

        Args:
          account_name: a str, the valid name of an account.
        Returns:
          A string, the name of the account.
        """
        # Check account name validity.
        if not self.account_regexp.match(account_name):
            # Note: This exception gets caught by BUILD_LEX() and converted into
            # a logged error.
            raise ValueError("Invalid account name: {}".format(account_name))

        # Reuse (intern) account strings as much as possible. This potentially
        # reduces memory usage a fair bit, because these strings are repeated
        # liberally.
        return self.accounts.setdefault(account_name, account_name)

    def CURRENCY(self, currency_name):
        """Process a CURRENCY token.

        Args:
          currency_name: the name of the currency.
        Returns:
          A new currency object; for now, these are simply represented
          as the currency name.
        """
        return currency_name

    def STRING(self, string):
        """Process a STRING token.

        Args:
          string: the string to process.
        Returns:
          The string. Nothing to be done or cleaned up. Eventually we might
          do some decoding here.
        """
        # If a multiline string, warm over a certain number of lines.
        if '\n' in string:
            num_lines = string.count('\n') + 1
            if num_lines > self.long_string_maxlines_default:
                # Note: This exception gets caught by BUILD_LEX() and converted
                # into a logged error.
                raise ValueError("String too long ({} lines)".format(num_lines))
        return string

    def NUMBER(self, number):
        """Process a NUMBER token. Convert into Decimal.

        Args:
          number: a str, the number to be converted.
        Returns:
          A Decimal instance built of the number string.
        """
        # Note: We don't use D() for efficiency here.
        # The lexer will only yield valid number strings.
        if ',' in number:
            # Check for a number with optional commas as thousands separator.
            #
            # Note: The regexp is liberally accepting commas in any position and
            # does not check for locale-specific placement of commas. This code
            # used to honor the user's locale by verifying that commas match
            # those the environment but we prefer to make parsing
            # locale-independent. An improvement would be to add an option to
            # specify the locale within Beancount itself and check numbers for
            # validity against that locale.
            if not self.number_regexp.match(number):
                # Note: This exception gets caught by BUILD_LEX() and converted
                # into a logged error.
                raise ValueError("Invalid number format: '{}'".format(number))
            # Remove commas.
            number = number.replace(',', '')
        return Decimal(number)

    def TAG(self, tag):
        """Process a TAG token.

        Args:
          tag: a str, the tag to be processed.
        Returns:
          The tag string itself. For now we don't need an object to represent
          those; keeping it simple.
        """
        return tag

    def LINK(self, link):
        """Process a LINK token.

        Args:
          link: a str, the name of the string.
        Returns:
          The link string itself. For now we don't need to represent this by
          an object.
        """
        return link

    def KEY(self, ident):
        """Process an identifier token.

        Args:
          ident: a str, the name of the key string.
        Returns:
          The link string itself. For now we don't need to represent this by
          an object.
        """
        return ident


def lex_iter(file, builder=None, encoding=None):
    """An iterator that yields all the tokens in the given file.

    Args:
      file: A string, the filename to run the lexer on, or a file object.
      builder: A builder of your choice. If not specified, a LexBuilder is
        used and discarded (along with its errors).
      encoding: A string (or None), the default encoding to use for strings.
    Yields:
      Tuples of the token (a string), the matched text (a string), and the line
      no (an integer).
    """
    # It would be more appropriate here to check for io.RawIOBase but
    # that does not work for io.BytesIO despite it implementing the
    # readinto() method.
    if not isinstance(file, io.IOBase):
        file = open(file, 'rb')
    if builder is None:
        builder = LexBuilder()
    parser = _parser.Parser(builder)
    yield from parser.lex(file, encoding=encoding)


def lex_iter_string(string, builder=None, encoding=None):
    """Parse an input string and print the tokens to an output file.

    Args:
      input_string: a str or bytes, the contents of the ledger to be parsed.
      builder: A builder of your choice. If not specified, a LexBuilder is
        used and discarded (along with its errors).
      encoding: A string (or None), the default encoding to use for strings.
    Returns:
      A iterator on the string. See lex_iter() for details.
    """
    if not isinstance(string, bytes):
        string = string.encode('utf8')
    file = io.BytesIO(string)
    yield from lex_iter(file, builder, encoding)



"""A plugin that asserts that all Commodity directives have a particular
attribute and that it is part of a set of enum values.

The configuration must be a mapping of attribute name to list of valid values,
like this:

    plugin "beancount.plugins.commodity_attr" "{
      'sector': ['Technology', 'Financials', 'Energy'],
      'name': None,
    }"

The plugin issues an error if a Commodity directive is missing the attribute, or
if the attribute value is not in the valid set. If you'd like to just ensure the
attribute is set, set the list of valid values to None, as in the 'name'
attribute in the example above.

"""
__copyright__ = "Copyright (C) 2018  Martin Blais"
__license__ = "GNU GPLv2"

import collections

from beancount.core import data

__plugins__ = ('validate_commodity_attr',)

ConfigError = collections.namedtuple('ConfigError', 'source message entry')
CommodityError = collections.namedtuple('CommodityError', 'source message entry')


def validate_commodity_attr(entries, unused_options_map, config_str):
    """Check that all Commodity directives have a valid attribute.

    Args:
      entries: A list of directives.
      unused_options_map: An options map.
      config_str: A configuration string.
    Returns:
      A list of new errors, if any were found.
    """
    errors = []

    # pylint: disable=eval-used
    config_obj = eval(config_str, {}, {})
    if not isinstance(config_obj, dict):
        errors.append(ConfigError(
            data.new_metadata('<commodity_attr>', 0),
            "Invalid configuration for commodity_attr plugin; skipping.", None))
        return entries, errors

    validmap = {attr: frozenset(values) if values is not None else None
                for attr, values in config_obj.items()}
    for entry in entries:
        if not isinstance(entry, data.Commodity):
            continue
        for attr, values in validmap.items():
            value = entry.meta.get(attr, None)
            if value is None:
                errors.append(CommodityError(
                    entry.meta,
                    "Missing attribute '{}' for Commodity directive {}".format(
                        attr, entry.currency), None))
                continue
            if values and value not in values:
                errors.append(CommodityError(
                    entry.meta,
                    "Invalid attribute '{}' for Commodity".format(value) +
                    " directive {}; valid options: {}".format(
                        entry.currency, ', '.join(values)), None))

    return entries, errors



"""Unit tests for treeify tool."""
__copyright__ = "Copyright (C) 2014-2017  Martin Blais"
__license__ = "GNU GPLv2"

import subprocess
import unittest
import textwrap


PROGRAM = __file__.replace('_test.py', '.py')
DEBUG = 0


def treeify(string, options=None):
    """Run treeify on the string.

    Args:
      string: The input string to feed treeify.
      options: An optional list of options for the subprogram.
    Returns:
      The treeified string.
    """
    pipe = subprocess.Popen([PROGRAM] + (options or []),
                            shell=False,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    output, errors = pipe.communicate(string.encode('utf-8'))
    return (pipe.returncode,
            output.decode('utf-8') if output else '',
            errors.decode('utf-8') if errors else '')


class TestTreeifyBase(unittest.TestCase):

    maxDiff = 8192

    def treeify(self, string, expect_errors=False, options=None):
        """Run treeify on the given string and assert no errors.

        Args:
          string: A string, the input contents to run on.
          expect_errors: A boolean, true if we should expect there to be errors.
          options: An optional list of options for the subprogram.
        Returns:
          The converted output string. This fails the test if there were any
          errors.
        """
        returncode, output, errors = treeify(string, options)
        actual_errors = returncode != 0 or bool(errors)
        if actual_errors != expect_errors:
            if expect_errors:
                self.fail("Missing expected errors")
            else:
                self.fail("Unexpected errors: {}".format(errors))
        return output

    def treeify_equal(self, string, expected, expect_errors=False, options=None):
        """Assert an expected treeification result.

        Args:
          string: An expected input contents string to treeify.
          expected: A string, the expected treeified output.
          expect_errors: A boolean, true if we should expect there to be errors.
          options: An optional list of options for the subprogram.
        Returns:
          The actual output string. This fails the test if there the expected
          output differed from the actual.
        """
        input_ = textwrap.dedent(string)
        output = self.treeify(input_, expect_errors, options)
        expected = textwrap.dedent(expected)
        if DEBUG:
            print('-(input)----------------------------------')
            print(input_)
            print('-(output)---------------------------------')
            print(output)
            print('-(expected)-------------------------------')
            print(expected)
            print('------------------------------------------')
        self.assertEqual(expected, output)
        return output


class TestTreeify(TestTreeifyBase):

    def test_simple(self):
        self.treeify_equal("""\
          2014-12-25 Assets:US:BofA:Checking                        5,545.01 USD
          2014-11-11 Assets:US:Federal:PreTax401k
          2014-10-04 Assets:US:Hooli:Vacation                         332.64 VACHR
          2014-11-07 Assets:US:Investment:Cash                     26,500.00 USD
          2014-12-15 Assets:US:Vanguard:Cash                           -0.07 USD
          2014-12-10 Assets:US:Vanguard:RGAGX                         174.22 RGAGX
          2014-10-19 Assets:US:Vanguard:VBMPX                         189.03 VBMPX
          2014-10-17 Equity:Opening-Balances                       -3,188.28 USD
          2014-12-12 Expenses:Food:Groceries                        6,483.71 USD
          2014-12-06 Expenses:Food:Restaurant                      10,990.74 USD
          2014-11-30 Expenses:Health:Dental:Insurance                 208.80 USD
          2014-11-09 Expenses:Health:Life:GroupTermLife             1,751.04 USD
          2014-12-07 Expenses:Health:Medical:Insurance              1,971.36 USD
          2014-10-12 Expenses:Health:Vision:Insurance               3,045.60 USD
          2014-12-28 Expenses:Home:Electricity                      2,080.00 USD
          2014-10-22 Expenses:Home:Internet                         2,560.22 USD
        """, """\
                     |-- Assets
                     |   `-- US
                     |       |-- BofA
          2014-12-25 |       |   `-- Checking                       5,545.01 USD
                     |       |-- Federal
          2014-11-11 |       |   `-- PreTax401k
                     |       |-- Hooli
          2014-10-04 |       |   `-- Vacation                         332.64 VACHR
                     |       |-- Investment
          2014-11-07 |       |   `-- Cash                          26,500.00 USD
                     |       `-- Vanguard
          2014-12-15 |           |-- Cash                              -0.07 USD
          2014-12-10 |           |-- RGAGX                            174.22 RGAGX
          2014-10-19 |           `-- VBMPX                            189.03 VBMPX
                     |-- Equity
          2014-10-17 |   `-- Opening-Balances                      -3,188.28 USD
                     `-- Expenses
                         |-- Food
          2014-12-12     |   |-- Groceries                          6,483.71 USD
          2014-12-06     |   `-- Restaurant                        10,990.74 USD
                         |-- Health
                         |   |-- Dental
          2014-11-30     |   |   `-- Insurance                        208.80 USD
                         |   |-- Life
          2014-11-09     |   |   `-- GroupTermLife                  1,751.04 USD
                         |   |-- Medical
          2014-12-07     |   |   `-- Insurance                      1,971.36 USD
                         |   `-- Vision
          2014-10-12     |       `-- Insurance                      3,045.60 USD
                         `-- Home
          2014-12-28         |-- Electricity                        2,080.00 USD
          2014-10-22         `-- Internet                           2,560.22 USD
        """)

    def test_empty_string(self):
        self.treeify_equal("""\
        """, """\
        """, True)

    def test_no_columns(self):
        self.treeify_equal("""\

          Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do
          eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad
          minim veniam, quis nostrud exercitation ullamco laboris nisi ut
          aliquip ex ea commodo consequat. Duis aute irure dolor in
          reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla
          pariatur. Excepteur sint occaecat cupidatat non proident, sunt in
          culpa qui officia deserunt mollit anim id est laborum.

        """, """\

          Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do
          eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad
          minim veniam, quis nostrud exercitation ullamco laboris nisi ut
          aliquip ex ea commodo consequat. Duis aute irure dolor in
          reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla
          pariatur. Excepteur sint occaecat cupidatat non proident, sunt in
          culpa qui officia deserunt mollit anim id est laborum.

        """, True)

    def test_flush_left(self):
        self.treeify_equal("""\
          Assets:US:Vanguard:Cash                           -0.07 USD
          Assets:US:Vanguard:RGAGX                         174.22 RGAGX
          Assets:US:Vanguard:VBMPX                         189.03 VBMPX
          Equity:Opening-Balances                       -3,188.28 USD
          Expenses:Food:Groceries                        6,483.71 USD
          Expenses:Food:Restaurant                      10,990.74 USD
          Expenses:Health:Dental:Insurance                 208.80 USD
          Expenses:Health:Life:GroupTermLife             1,751.04 USD
        """, """\
          |-- Assets
          |   `-- US
          |       `-- Vanguard
          |           |-- Cash                              -0.07 USD
          |           |-- RGAGX                            174.22 RGAGX
          |           `-- VBMPX                            189.03 VBMPX
          |-- Equity
          |   `-- Opening-Balances                      -3,188.28 USD
          `-- Expenses
              |-- Food
              |   |-- Groceries                          6,483.71 USD
              |   `-- Restaurant                        10,990.74 USD
              `-- Health
                  |-- Dental
                  |   `-- Insurance                        208.80 USD
                  `-- Life
                      `-- GroupTermLife                  1,751.04 USD
        """, False)

    def test_flush_right(self):
        self.treeify_equal("""\
          2014-01-01 Assets:US:Vanguard:Cash
          2014-01-01 Assets:US:Vanguard:RGAGX
          2014-01-01 Assets:US:Vanguard:VBMPX
          2014-01-01 Equity:Opening-Balances
          2014-01-01 Expenses:Food:Groceries
          2014-01-01 Expenses:Food:Restaurant
          2014-01-01 Expenses:Health:Dental:Insurance
          2014-01-01 Expenses:Health:Life:GroupTermLife
        """, """\
                     |-- Assets
                     |   `-- US
                     |       `-- Vanguard
          2014-01-01 |           |-- Cash
          2014-01-01 |           |-- RGAGX
          2014-01-01 |           `-- VBMPX
                     |-- Equity
          2014-01-01 |   `-- Opening-Balances
                     `-- Expenses
                         |-- Food
          2014-01-01     |   |-- Groceries
          2014-01-01     |   `-- Restaurant
                         `-- Health
                             |-- Dental
          2014-01-01         |   `-- Insurance
                             `-- Life
          2014-01-01             `-- GroupTermLife
        """, False)

    def test_two_columns(self):
        self.treeify_equal("""\
          2014-01-01 Assets:US:Vanguard:Cash              | Assets:US:Vanguard:Cash            XX
          2014-01-01 Assets:US:Vanguard:RGAGX             | Assets:US:Vanguard:RGAGX           XX
          2014-01-01 Assets:US:Vanguard:VBMPX             | Assets:US:Vanguard:VBMPX           XX
          2014-01-01 Equity:Opening-Balances              | Equity:Opening-Balances            XX
          2014-01-01 Expenses:Food:Groceries              | Expenses:Food:Groceries            XX
          2014-01-01 Expenses:Food:Restaurant             | Expenses:Food:Restaurant           XX
          2014-01-01 Expenses:Health:Dental:Insurance     | Expenses:Health:Dental:Insurance   XX
          2014-01-01 Expenses:Health:Life:GroupTermLife   | Expenses:Health:Life:GroupTermLife XX
        """, """\
                     |-- Assets
                     |   `-- US
                     |       `-- Vanguard
          2014-01-01 |           |-- Cash                 | Assets:US:Vanguard:Cash            XX
          2014-01-01 |           |-- RGAGX                | Assets:US:Vanguard:RGAGX           XX
          2014-01-01 |           `-- VBMPX                | Assets:US:Vanguard:VBMPX           XX
                     |-- Equity
          2014-01-01 |   `-- Opening-Balances             | Equity:Opening-Balances            XX
                     `-- Expenses
                         |-- Food
          2014-01-01     |   |-- Groceries                | Expenses:Food:Groceries            XX
          2014-01-01     |   `-- Restaurant               | Expenses:Food:Restaurant           XX
                         `-- Health
                             |-- Dental
          2014-01-01         |   `-- Insurance            | Expenses:Health:Dental:Insurance   XX
                             `-- Life
          2014-01-01             `-- GroupTermLife        | Expenses:Health:Life:GroupTermLife XX
        """, False)

    def test_overlapping_column(self):
        self.treeify_equal("""\
          Assets:US:Vanguard:Cash            100.00 USD
          Assets:US:Vanguard:RGAGX           100.00 USD
          Assets:US:Vanguard:VBMPX           100.00 USD
          Equity:Opening-Balances          100.00 USD <- overlapping
          Expenses:Food:Groceries            100.00 USD
          Expenses:Food:Restaurant           100.00 USD
          Expenses:Health:Dental:Insurance   100.00 USD
          Expenses:Health:Life:GroupTermLife 100.00 USD
        """, """\
          Assets:US:Vanguard:Cash            100.00 USD
          Assets:US:Vanguard:RGAGX           100.00 USD
          Assets:US:Vanguard:VBMPX           100.00 USD
          Equity:Opening-Balances          100.00 USD <- overlapping
          Expenses:Food:Groceries            100.00 USD
          Expenses:Food:Restaurant           100.00 USD
          Expenses:Health:Dental:Insurance   100.00 USD
          Expenses:Health:Life:GroupTermLife 100.00 USD
        """, True)

    def test_parents(self):
        self.treeify_equal("""\
          Assets:US                          100.00 USD
          Assets:US:Vanguard                 101.00 USD
          Assets:US:Vanguard:VBMPX           102.00 USD
        """, """\
          `-- Assets
              `-- US                         100.00 USD
                  `-- Vanguard               101.00 USD
                      `-- VBMPX              102.00 USD
        """, False)

    def test_unsorted(self):
        # Check when the same parent comes around multiple times (the column
        # values aren't sorted properly).
        self.treeify_equal("""\
          Assets:US:Vanguard:VBMPX           100.00 USD
          Assets:CA:Bank:Checking            101.00 USD
          Assets:US:Vanguard:VBMPX           102.00 USD
          Assets:ES:Caja                     103.00 USD
          Assets:US:Vanguard:VBMPX           104.00 USD
        """, """\
         `-- Assets
             |-- US
             |   `-- Vanguard
             |       `-- VBMPX              100.00 USD
             |-- CA
             |   `-- Bank
             |       `-- Checking           101.00 USD
             |-- US
             |   `-- Vanguard
             |       `-- VBMPX              102.00 USD
             |-- ES
             |   `-- Caja                   103.00 USD
             `-- US
                 `-- Vanguard
                     `-- VBMPX              104.00 USD
        """, False)

    def test_consecutive(self):
        self.treeify_equal("""\
          Assets:US:Vanguard:VBMPX           100.00 USD
          Assets:US:Vanguard:VBMPX           100.00 USD
          Assets:US:Vanguard:VBMPX           102.00 USD
          Assets:US:Vanguard:VBMPX           104.00 USD
        """, """\
          `-- Assets
              `-- US
                  `-- Vanguard
                      `-- VBMPX              100.00 USD
                                             100.00 USD
                                             102.00 USD
                                             104.00 USD
        """, False)

    def test_noise_before(self):
        self.treeify_equal("""\
          Account                         Balance
          ---------------------------------------------
          Assets:US:Vanguard:VBMPX     100.00 USD
          Assets:US:Vanguard:VBMPX     101.00 USD
          Expenses:Food:Groceries      102.00 USD
        """, """\
          Account                         Balance
          ---------------------------------------------
          |-- Assets
          |   `-- US
          |       `-- Vanguard
          |           `-- VBMPX        100.00 USD
          |                            101.00 USD
          `-- Expenses
              `-- Food
                  `-- Groceries        102.00 USD
        """, False)

    def test_noise_after(self):
        self.treeify_equal("""\
          Assets:US:Vanguard:VBMPX     100.00 USD
          Assets:US:Vanguard:VBMPX     101.00 USD
          Expenses:Food:Groceries      102.00 USD
          >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
          Not matching: yep, not matching.
        """, """\
          |-- Assets
          |   `-- US
          |       `-- Vanguard
          |           `-- VBMPX        100.00 USD
          |                            101.00 USD
          `-- Expenses
              `-- Food
                  `-- Groceries        102.00 USD
          >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
          Not matching: yep, not matching.
        """, False)

    def test_noise_middle_between_nodes(self):
        self.treeify_equal("""\
          Assets:US:Vanguard:VBMPX     100.00 USD
          Assets:US:Vanguard:VBMPX     101.00 USD
          >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
          Expenses:Food:Groceries      102.00 USD
        """, """\
          |-- Assets
          |   `-- US
          |       `-- Vanguard
          |           `-- VBMPX        100.00 USD
          |                            101.00 USD
          >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
          `-- Expenses
              `-- Food
                  `-- Groceries        102.00 USD
        """, False)

    def test_noise_middle_same_node(self):
        self.treeify_equal("""\
          Assets:US:Vanguard:VBMPX     100.00 USD
          >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
          Assets:US:Vanguard:VBMPX     101.00 USD
          Expenses:Food:Groceries      102.00 USD
        """, """\
          |-- Assets
          |   `-- US
          |       `-- Vanguard
          |           `-- VBMPX        100.00 USD
          >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
          |                            101.00 USD
          `-- Expenses
              `-- Food
                  `-- Groceries        102.00 USD
        """, False)

    def test_noise_middle_parent_child(self):
        self.treeify_equal("""\
          Assets:US:Vanguard           100.00 USD
          >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
          Assets:US:Vanguard:VBMPX     101.00 USD
          Expenses:Food:Groceries      102.00 USD
        """, """\
          |-- Assets
          |   `-- US
          |       `-- Vanguard         100.00 USD
          >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
          |           `-- VBMPX        101.00 USD
          `-- Expenses
              `-- Food
                  `-- Groceries        102.00 USD
        """, False)


    def test_filenames(self):
        self.treeify_equal("""\
          -rwxr-xr-x    1 blais   5000  44708 Sep 27 13:54 ./beancount/scripts/example.py
          -rw-r--r--    1 blais   5000    678 Sep 21 12:11 ./beancount/scripts/example_test.py
          -rwxr-xr-x    1 blais   5000   3321 Sep 23 14:30 ./beancount/scripts/format.py
          -rw-r--r--    1 blais   5000   1341 Sep 23 14:30 ./beancount/scripts/format_test.py
          -rwxr-xr-x    1 blais   5000   8714 Sep 20 13:09 ./beancount/scripts/query.py
          -rw-r--r--    1 blais   5000   4796 Sep  3 07:26 ./beancount/scripts/query_test.py
          drwxr-xr-x   18 blais   5000    612 Sep 29 22:18 ./beancount/utils
          -rw-r--r--    1 blais   5000     48 Jun 15 11:53 ./beancount/utils/__init__.py
          -rw-r--r--    1 blais   5000   1464 Sep  3 07:26 ./beancount/utils/bisect_key.py
          -rw-r--r--    1 blais   5000   1330 Sep  3 07:26 ./beancount/utils/bisect_key_test.py
          -rw-r--r--    1 blais   5000   1203 Aug 17 13:06 ./beancount/utils/file_utils.py
          -rw-r--r--    1 blais   5000   1130 Aug 17 13:06 ./beancount/utils/file_utils_test.py
          -rw-r--r--    1 blais   5000   8377 Sep 15 06:53 ./beancount/utils/misc_utils.py
          -rw-r--r--    1 blais   5000   4783 Sep 13 15:17 ./beancount/utils/misc_utils_test.py
        """, """\
                                                           `-- .
                                                               `-- beancount
                                                                   |-- scripts
          -rwxr-xr-x    1 blais   5000  44708 Sep 27 13:54         |   |-- example.py
          -rw-r--r--    1 blais   5000    678 Sep 21 12:11         |   |-- example_test.py
          -rwxr-xr-x    1 blais   5000   3321 Sep 23 14:30         |   |-- format.py
          -rw-r--r--    1 blais   5000   1341 Sep 23 14:30         |   |-- format_test.py
          -rwxr-xr-x    1 blais   5000   8714 Sep 20 13:09         |   |-- query.py
          -rw-r--r--    1 blais   5000   4796 Sep  3 07:26         |   `-- query_test.py
          drwxr-xr-x   18 blais   5000    612 Sep 29 22:18         `-- utils
          -rw-r--r--    1 blais   5000     48 Jun 15 11:53             |-- __init__.py
          -rw-r--r--    1 blais   5000   1464 Sep  3 07:26             |-- bisect_key.py
          -rw-r--r--    1 blais   5000   1330 Sep  3 07:26             |-- bisect_key_test.py
          -rw-r--r--    1 blais   5000   1203 Aug 17 13:06             |-- file_utils.py
          -rw-r--r--    1 blais   5000   1130 Aug 17 13:06             |-- file_utils_test.py
          -rw-r--r--    1 blais   5000   8377 Sep 15 06:53             |-- misc_utils.py
          -rw-r--r--    1 blais   5000   4783 Sep 13 15:17             `-- misc_utils_test.py
        """, False, options=['--pattern=([^ ]+)(/[^ ]+)+', '--split=/'])

    def test_filenames_tree(self):
        self.treeify_equal("""\
          -rwxr-xr-x    1 blais   5000  44708 Sep 27 13:54 ./beancount
          -rw-r--r--    1 blais   5000    678 Sep 21 12:11 ./beancount/scripts
          -rwxr-xr-x    1 blais   5000   3321 Sep 23 14:30 ./beancount/scripts/format.py
        """, """\
                                                           `-- .
          -rwxr-xr-x    1 blais   5000  44708 Sep 27 13:54     `-- beancount
          -rw-r--r--    1 blais   5000    678 Sep 21 12:11         `-- scripts
          -rwxr-xr-x    1 blais   5000   3321 Sep 23 14:30             `-- format.py
        """, False, options=['--pattern=([^ ]+)(/[^ ]+)+', '--split=/'])

    def test_width_wider(self):
        # The treeified column should be wider as it needs to.
        self.treeify_equal("""\
          Equity:B       100.00 USD
          Equity:B:C     101.00 USD
          Equity:B:C:D   102.00 USD
        """, """\
          `-- Equity
              `-- B           100.00 USD
                  `-- C       101.00 USD
                      `-- D   102.00 USD
        """, False)

    def test_width_narrower(self):
        # The treeified column should be of the same width even though it does
        # not need to.
        self.treeify_equal("""\
          Assets:Bcdefg                100.00 USD
          Assets:Bcdefg:Cdefgh         101.00 USD
          Assets:Bcdefg:Cdefgh:Defghij 102.00 USD
        """, """\
          `-- Assets
              `-- Bcdefg               100.00 USD
                  `-- Cdefgh           101.00 USD
                      `-- Defghij      102.00 USD
        """, False)

    def test_whitespace(self):
        # Test with columns that have whitespace in them.
        self.treeify_equal(
            "Assets:Bank:DE:Bankname:Tagesgeld         3.57 EUR\n"
            "Equity:Opening-Balances                  -3.57 EUR\n"
            "Expenses                          \n"
            "Income                            \n"
            "Liabilities                       \n",
            """\
             |-- Assets
             |   `-- Bank
             |       `-- DE
             |           `-- Bankname
             |               `-- Tagesgeld             3.57 EUR
             |-- Equity
             |   `-- Opening-Balances                 -3.57 EUR
             |-- Expenses
             |-- Income
             `-- Liabilities
        """, False)


if __name__ == '__main__':
    unittest.main()



"""Write output files for the tutorial commands.
"""
__copyright__ = "Copyright (C) 2014, 2016-2017  Martin Blais"
__license__ = "GNU GPLv2"

import logging
from os import path

from beancount.utils import test_utils
from beancount.parser import version
from beancount.reports import report


# pylint: disable=bad-whitespace
COMMANDS = [
    ('balances'                 , "bean-report {} balances"),
    ('help-reports'             , "bean-report --help-reports"),
    ('help-subcmd'              , "bean-report {} balances --help"),
    ('help-global'              , "bean-report --help"),
    ('help-formats'             , "bean-report --help-formats"),
    ('balances-restrict'        , "bean-report {} balances -e ETrade"),
    ('balances-restrict-cost'   , "bean-report {} balances -e ETrade -c"),
    ('balances-tree'            , "bean-report {} balances"),  # TODO(blais): Add treeify.
    ('balsheet'                 , "bean-report {} balsheet"),
    ('journal'                  , ("bean-report {} journal -w 120 "
                                   "-a Assets:US:BofA:Checking")),
    ('journal-with-balance'     , ("bean-report {} journal -w 120 "
                                   "-a Assets:US:BofA:Checking -b")),
    ('invest'                   , ("bean-report {} journal -w 120 "
                                   "-a Assets:US:ETrade:GLD -b")),
    ('invest-with-cost'         , ("bean-report {} journal -w 120 "
                                   "-a Assets:US:ETrade:GLD -b -c")),
    ('journal-unrestricted'     , "bean-report {} journal -w 120 -b"),
    ('holdings'                 , "bean-report {} holdings"),
    ('holdings-by-account'      , "bean-report {} holdings --by account"),
    ('holdings-by-root-account' , "bean-report {} holdings --by root-account"),
    ('holdings-by-commodity'    , "bean-report {} holdings --by commodity"),
    ('holdings-by-currency'     , "bean-report {} holdings --by currency"),
    ('networth'                 , "bean-report {} networth"),
    ('accounts'                 , "bean-report {} accounts"),
    ('events'                   , "bean-report {} events"),
    ('stats-directives'         , "bean-report {} stats-directives"),

    ('stats-postings'           , "bean-report {} stats-postings"),
    ('holdings-csv'             , "bean-report -f csv {} holdings"),
    ]


def main():
    logging.basicConfig(level=logging.INFO, format='%(levelname)-8s: %(message)s')
    parser = version.ArgumentParser(description=__doc__.strip())
    parser.add_argument('filename', help='Beancount filename')
    parser.add_argument('output_directory', help='Output directory for the tutorial files')
    args = parser.parse_args()
    rootdir = test_utils.find_repository_root(__file__)

    for report_name, command_template in COMMANDS:
        logging.info('Generating %s: %s', report_name, command_template)
        rootdir = test_utils.find_repository_root(__file__)
        output_filename = path.join(args.output_directory, '{}.output'.format(report_name))
        errors_filename = path.join(args.output_directory, '{}.errors'.format(report_name))
        command = command_template.format(args.filename)
        command_args = command.split()[1:]
        with test_utils.capture('stdout') as stdout:
            with test_utils.capture('stderr') as stderr:
                try:
                    returncode = report.main(command_args)
                except SystemExit as exc:
                    returncode = exc.code
                if returncode != 0:
                    with open(errors_filename) as efile:
                        errors = efile.read()
                    raise RuntimeError(
                        "Error running '{}': exit with {}; errors: {}".format(
                            command, returncode, errors))

        # NOTE(blais): This could be removed by improving the capture() function.
        output_filename = path.join(args.output_directory,
                                    '{}.output'.format(report_name))
        with open(output_filename, 'w') as outfile:
            outfile.write(stdout.getvalue())

        if stderr.getvalue():
            errors_filename = path.join(args.output_directory,
                                        '{}.errors'.format(report_name))
            with open(errors_filename, 'w') as outfile:
                outfile.write(stderr.getvalue())

    return 0


if __name__ == '__main__':
    main()



"""This module has code that can build a database of historical prices at
various times, from which unrealized capital gains and market value can be
deduced.

Prices are deduced from Price entries found in the file, or perhaps
created by scripts (for example you could build a script that will fetch
live prices online and create entries on-the-fly).
"""
__copyright__ = "Copyright (C) 2013-2017  Martin Blais"
__license__ = "GNU GPLv2"

import collections

from beancount.core.number import ONE
from beancount.core.number import ZERO
from beancount.core.data import Price
from beancount.core import data
from beancount.utils import misc_utils
from beancount.utils import bisect_key


def get_last_price_entries(entries, date):
    """Run through the entries until the given date and return the last
    Price entry encountered for each (currency, cost-currency) pair.

    Args:
      entries: A list of directives.
      date: An instance of datetime.date. If None, the very latest price
        is returned.
    Returns:
      A list of price entries.
    """
    price_entry_map = {}
    for entry in entries:
        if date is not None and entry.date >= date:
            break
        if isinstance(entry, Price):
            base_quote = (entry.currency, entry.amount.currency)
            price_entry_map[base_quote] = entry
    return sorted(price_entry_map.values(), key=data.entry_sortkey)


class PriceMap(dict):
    """A price map dictionary.

    The keys include both the set of forward (base, quote) pairs and their
    inverse. In order to determine which are the forward pairs, access the
    'forward_pairs' attribute

    Attributes:
      forward_pairs: A list of (base, quote) keys for the forward pairs.
    """
    __slots__ = ('forward_pairs',)


def build_price_map(entries):
    """Build a price map from a list of arbitrary entries.

    If multiple prices are found for the same (currency, cost-currency) pair at
    the same date, the latest date is kept and the earlier ones (for that day)
    are discarded.

    If inverse price pairs are found, e.g. USD in AUD and AUD in USD, the
    inverse that has the smallest number of price points is converted into the
    one that has the most price points. In that way they are reconciled into a
    single one.

    Args:
      entries: A list of directives, hopefully including some Price and/or
      Transaction entries.
    Returns:
      A dict of (currency, cost-currency) keys to sorted lists of (date, number)
      pairs, where 'date' is the date the price occurs at and 'number' a Decimal
      that represents the price, or rate, between these two
      currencies/commodities. Each date occurs only once in the sorted list of
      prices of a particular key. All of the inverses are automatically
      generated in the price map.
    """
    # Fetch a list of all the price entries seen in the ledger.
    price_entries = [entry
                     for entry in entries
                     if isinstance(entry, Price)]

    # Build a map of exchange rates between these units.
    # (base-currency, quote-currency) -> List of (date, rate).
    price_map = collections.defaultdict(list)
    for price in price_entries:
        base_quote = (price.currency, price.amount.currency)
        price_map[base_quote].append((price.date, price.amount.number))

    # Find pairs of inversed units.
    inversed_units = []
    for base_quote, values in price_map.items():
        base, quote = base_quote
        if (quote, base) in price_map:
            inversed_units.append(base_quote)

    # Find pairs of inversed units, and swallow the conversion with the smaller
    # number of rates into the other one.
    for base, quote in inversed_units:
        bq_prices = price_map[(base, quote)]
        qb_prices = price_map[(quote, base)]
        remove = ((base, quote)
                  if len(bq_prices) < len(qb_prices)
                  else (quote, base))
        base, quote = remove

        remove_list = price_map[remove]
        insert_list = price_map[(quote, base)]
        del price_map[remove]

        inverted_list = [(date, ONE/rate)
                         for (date, rate) in remove_list
                         if rate != ZERO]
        insert_list.extend(inverted_list)

    # Unzip and sort each of the entries and eliminate duplicates on the date.
    sorted_price_map = PriceMap({
        base_quote: list(misc_utils.sorted_uniquify(date_rates, lambda x: x[0], last=True))
        for (base_quote, date_rates) in price_map.items()})

    # Compute and insert all the inverted rates.
    forward_pairs = list(sorted_price_map.keys())
    for (base, quote), price_list in list(sorted_price_map.items()):
        # Note: You have to filter out zero prices for zero-cost postings, like
        # gifted options.
        sorted_price_map[(quote, base)] = [
            (date, ONE/price) for date, price in price_list
            if price != ZERO]

    sorted_price_map.forward_pairs = forward_pairs
    return sorted_price_map


def normalize_base_quote(base_quote):
    """Convert a slash-separated string to a pair of strings.

    Args:
      base_quote: A pair of strings, the base currency to lookup, and the quote
        currency to lookup, which expresses which units the base currency is
        denominated in. This may also just be a string, with a '/' separator.
    Returns:
      A pair of strings.
    """
    if isinstance(base_quote, str):
        base_quote_norm = tuple(base_quote.split('/'))
        assert len(base_quote_norm) == 2, base_quote
        base_quote = base_quote_norm
    assert isinstance(base_quote, tuple), base_quote
    return base_quote


def _lookup_price_and_inverse(price_map, base_quote):
    """Lookup the (base, quote) tuple in the price map and its inverse.
    If not found, raise an appropriate exception.

    Note: this is meant to be an INTERNAL helper function, use the get_*
    functions to obtain values from a price_map object.

    Args:
      price_map: A price map, which is a dict of (base, quote) -> list of (date,
        number) tuples, as created by build_price_map.
      base_quote: A pair of strings, (base, quote) currencies.
        No normalization is done.
    Returns:
      A list of price-dates, if successful.
    Raises:
      KeyError: If the base_quote and its inverse both weren't able to be looked
        up.
    """
    try:
        return price_map[base_quote]
    except KeyError as exc:
        base, quote = base_quote
        prices = price_map.get((quote, base), None)
        if prices:
            return prices
        else:
            raise


def get_all_prices(price_map, base_quote):
    """Return a sorted list of all (date, number) price pairs.

    Args:
      price_map: A price map, which is a dict of (base, quote) -> list of (date,
        number) tuples, as created by build_price_map.
      base_quote: A pair of strings, the base currency to lookup, and the quote
        currency to lookup, which expresses which units the base currency is
        denominated in. This may also just be a string, with a '/' separator.
    Returns:
      A list of (date, Decimal) pairs, sorted by date.
    Raises:
      KeyError: If the base/quote could not be found.
    """
    base_quote = normalize_base_quote(base_quote)
    return _lookup_price_and_inverse(price_map, base_quote)


def get_latest_price(price_map, base_quote):
    """Return the latest price/rate from a price map for the given base/quote pair.
    This is often used to just get the 'current' price if you're looking at the
    entire set of entries.

    Args:
      price_map: A price map, which is a dict of (base, quote) -> list of (date,
        number) tuples, as created by build_price_map.
    Returns:
      A pair of (date, number), where 'date' is a datetime.date instance and
      'number' is a Decimal of the price, or rate, at that date. The date is the
      latest date which we have an available price for in the price map.
    """
    base_quote = normalize_base_quote(base_quote)

    # Handle the degenerate case of a currency priced into its own.
    base, quote = base_quote
    if quote is None or base == quote:
        return (None, ONE)

    # Look up the list and return the latest element. The lists are assumed to
    # be sorted.
    try:
        price_list = _lookup_price_and_inverse(price_map, base_quote)
    except KeyError:
        price_list = None
    if price_list:
        return price_list[-1]
    else:
        return None, None


def get_price(price_map, base_quote, date=None):
    """Return the price as of the given date.

    If the date is unspecified, return the latest price.

    Args:
      price_map: A price map, which is a dict of (base, quote) -> list of (date,
        number) tuples, as created by build_price_map.
      base_quote: A pair of strings, the base currency to lookup, and the quote
        currency to lookup, which expresses which units the base currency is
        denominated in. This may also just be a string, with a '/' separator.
      date: A datetime.date instance, the date at which we want the conversion
        rate.
    Returns:
      A pair of (datetime.date, Decimal) instance. If no price information could
      be found, return (None, None).
    """
    if date is None:
        return get_latest_price(price_map, base_quote)

    base_quote = normalize_base_quote(base_quote)

    # Handle the degenerate case of a currency priced into its own.
    base, quote = base_quote
    if quote is None or base == quote:
        return (None, ONE)

    try:
        price_list = _lookup_price_and_inverse(price_map, base_quote)
        index = bisect_key.bisect_right_with_key(price_list, date, key=lambda x: x[0])
        if index == 0:
            return None, None
        else:
            return price_list[index-1]
    except KeyError:
        return None, None
